# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
A script to plot phonon band structure diagrams.

TODO:
 * automatically plot dos if present in band.yaml
 * make band structure from vasprun displacement/DFPT files
 * deal with magnetic moments
 * Read force_constants.hdf5
 * read settings from phonopy config file
 * prefix file names
"""

import argparse
import json
import logging
import os
import sys
import warnings
from os.path import isfile

import numpy as np

warnings.filterwarnings("ignore", category=FutureWarning, module="h5py")
import matplotlib as mpl

mpl.use("Agg")
from matplotlib import rcParams
from phonopy.units import CmToEv, THzToCm, VaspToCm, VaspToEv, VaspToTHz
from pymatgen.io.phonopy import get_ph_bs_symm_line
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine

from sumo.io.castep import CastepPhonon
from sumo.phonon.phonopy import load_phonopy
from sumo.plotting.phonon_bs_plotter import SPhononBSPlotter
from sumo.symmetry.kpoints import get_path_data

__author__ = "Alex Ganose"
__version__ = "1.0"
__maintainer__ = "Alex Ganose"
__email__ = "alexganose@googlemail.com"
__date__ = "Jan 17, 2018"


def phonon_bandplot(
    filename,
    poscar=None,
    prefix=None,
    directory=None,
    dim=None,
    born=None,
    qmesh=None,
    spg=None,
    primitive_axis=None,
    line_density=60,
    units="THz",
    symprec=0.01,
    mode="bradcrack",
    kpt_list=None,
    eigenvectors=None,
    labels=None,
    height=6.0,
    width=6.0,
    style=None,
    no_base_style=False,
    ymin=None,
    ymax=None,
    image_format="pdf",
    dpi=400,
    plt=None,
    fonts=None,
    dos=None,
    to_json=None,
    from_json=None,
    to_web=None,
    legend=None,
):
    """A script to plot phonon band structure diagrams.

    Args:
        filename (str): Path to phonopy output. Can be a band structure yaml
            file, ``FORCE_SETS``, ``FORCE_CONSTANTS``, or
            ``force_constants.hdf5``.
        poscar (:obj:`str`, optional): Path to POSCAR file of unitcell. Not
            required if plotting the phonon band structure from a yaml file. If
            not specified, the script will search for a POSCAR file in the
            current directory.
        prefix (:obj:`str`, optional): Prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
        dim (:obj:`list`, optional): supercell matrix
        born (:obj:`str`, optional): Path to file containing Born effective
            charges. Should be in the same format as the file produced by the
            ``phonopy-vasp-born`` script provided by phonopy.
        qmesh (:obj:`list` of :obj:`int`, optional): Q-point mesh to use for
            calculating the density of state. Formatted as a 3x1 :obj:`list` of
            :obj:`int`.
        spg (:obj:`str` or :obj:`int`, optional): The space group international
            number or symbol to override the symmetry determined by spglib.
            This is not recommended and only provided for testing purposes.
            This option will only take effect when ``mode = 'bradcrack'``.
        primitive_axis (:obj:`list` or :obj:`str`, optional): The
            transformation matrix from the conventional to primitive cell. Only
            required when the conventional cell was used as the starting
            structure. Should be provided as a 3x3 :obj:`list` of
            :obj:`float`. Alternatively the string 'auto' may be used to
            request that a primitive matrix is determined automatically.
        line_density (:obj:`int`, optional): Density of k-points along the
            path.
        units (:obj:`str`, optional): Units of phonon frequency. Accepted
            (case-insensitive) values are Thz, cm-1, eV, meV.
        symprec (:obj:`float`, optional): Tolerance for space-group-finding
            operations
        mode (:obj:`str`, optional): Method used for calculating the
            high-symmetry path. The options are:

            bradcrack
                Use the paths from Bradley and Cracknell. See [brad]_.

            pymatgen
                Use the paths from pymatgen. See [curt]_.

            seekpath
                Use the paths from SeeK-path. See [seek]_.

        kpt_list (:obj:`list`, optional): List of k-points to use, formatted as
            a list of subpaths, each containing a list of fractional k-points.
            For example::

                [ [[0., 0., 0.], [0., 0., 0.5]],
                  [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]

            Will return points along ``0 0 0 -> 0 0 1/2 | 1/2 0 0
            -> 1/2 1/2 0``
        labels (:obj:`list`, optional): The k-point labels. These should
            be provided as a :obj:`list` of :obj:`str` for each subpath of the
            overall path. For example::

                [ ['Gamma', 'Z'], ['X', 'M'] ]

            combined with the above example for ``kpt_list`` would indicate the
            path: Gamma -> Z | X -> M. If no labels are provided, letters from
            A -> Z will be used instead.
        eigenvectors (:obj:`bool`, optional): Write the eigenvectors to the
            yaml file. (Always True if to_web is set.)
        dos (str): Path to Phonopy total dos .dat file
        height (:obj:`float`, optional): The height of the plot.
        width (:obj:`float`, optional): The width of the plot.
        ymin (:obj:`float`, optional): The minimum energy on the y-axis.
        ymax (:obj:`float`, optional): The maximum energy on the y-axis.
        style (:obj:`list` or :obj:`str`, optional): (List of) matplotlib style
            specifications, to be composed on top of Sumo base style.
        no_base_style (:obj:`bool`, optional): Prevent use of sumo base style.
            This can make alternative styles behave more predictably.
        image_format (:obj:`str`, optional): The image file format. Can be any
            format supported by matplotlib, including: png, jpg, pdf, and svg.
            Defaults to pdf.
        dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
            the image.
        plt (:obj:`matplotlib.pyplot`, optional): A
            :obj:`matplotlib.pyplot` object to use for plotting.
        fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
            a single font, specified as a :obj:`str`, or several fonts,
            specified as a :obj:`list` of :obj:`str`.
        to_json (:obj:`str`, optional): JSON file to dump the Pymatgen band
            path object.
        from_json (:obj:`list` or :obj:`str`, optional): (List of) JSON
            bandpath data filename(s) to import and overlay.
        to_web (:obj:`str`, optional): JSON file to write data for
            visualisation with http://henriquemiranda.github.io/phononwebsite
        legend (:obj:`list` or :obj:`None`, optional): Legend labels. If None,
            don't add a legend. With a list length equal to from_json, label
            json inputs only. With one extra list entry, label all lines
            beginning with new plot.

    Returns:
        A matplotlib pyplot object.
    """
    save_files = False if plt else True  # don't save if pyplot object provided
    if isinstance(from_json, str):
        from_json = [from_json]

    if eigenvectors is None:
        if to_web is None:
            eigenvectors = False
        else:
            eigenvectors = True
    elif not eigenvectors and to_web is not None:
        raise ValueError("Cannot set eigenvectors=False and write web JSON")

    if filename is None:
        if from_json is None:
            filename = "FORCE_SETS"
            bs, phonon = _bs_from_filename(
                filename,
                poscar,
                dim,
                symprec,
                spg,
                kpt_list,
                labels,
                primitive_axis,
                units,
                born,
                mode,
                eigenvectors,
                line_density,
            )

        else:
            logging.info(f"No input data, using file {from_json[0]} to construct plot")
            with open(from_json[0]) as f:
                bs = PhononBandStructureSymmLine.from_dict(json.load(f))
            from_json = from_json[1:]
            phonon = None
    else:
        bs, phonon = _bs_from_filename(
            filename,
            poscar,
            dim,
            symprec,
            spg,
            kpt_list,
            labels,
            primitive_axis,
            units,
            born,
            mode,
            eigenvectors,
            line_density,
        )

    if to_json is not None:
        logging.info(f"Writing symmetry lines to {to_json}")
        with open(to_json, "wt") as f:
            f.write(bs.to_json())

    if to_web is not None:
        logging.info(f"Writing visualisation JSON to {to_web}")
        bs.write_phononwebsite(to_web)

    # Replace dos filename with data array
    if dos is not None:
        if phonon is None:
            logging.error("Cannot use phonon DOS without Phonopy")
            sys.exit()

        if isfile(dos):
            dos = np.genfromtxt(dos, comments="#")
        elif dos:
            phonon.set_mesh(
                qmesh,
                is_gamma_center=False,
                is_eigenvectors=True,
                is_mesh_symmetry=False,
            )
            phonon.set_total_DOS()
            dos_freq, dos_val = phonon.get_total_DOS()
            dos = np.zeros((len(dos_freq), 2))
            dos[:, 0], dos[:, 1] = dos_freq, dos_val

    plotter = SPhononBSPlotter(bs)
    plt = plotter.get_plot(
        units=units,
        ymin=ymin,
        ymax=ymax,
        height=height,
        width=width,
        plt=plt,
        fonts=fonts,
        dos=dos,
        style=style,
        no_base_style=no_base_style,
        from_json=from_json,
        legend=legend,
    )

    if save_files:
        basename = f"phonon_band.{image_format}"
        filename = f"{prefix}_{basename}" if prefix else basename

        if directory:
            filename = os.path.join(directory, filename)

        if dpi is None:
            dpi = rcParams["figure.dpi"]
        plt.savefig(filename, format=image_format, dpi=dpi, bbox_inches="tight")

        filename = save_data_files(bs, prefix=prefix, directory=directory)
        return filename
    else:
        return plt


def save_data_files(bs, prefix=None, directory=None):
    """Write the phonon band structure data files to disk.

    Args:
        bs (:obj:`~pymatgen.phonon.bandstructure.PhononBandStructureSymmLine`):
            The phonon band structure.
        prefix (:obj:`str`, optional): Prefix for data file.
        directory (:obj:`str`, optional): Directory in which to save the data.

    Returns:
        str: The filename of the written data file.
    """
    filename = "phonon_band.dat"
    filename = f"{prefix}_phonon_band.dat" if prefix else filename
    directory = directory if directory else "."
    filename = os.path.join(directory, filename)

    with open(filename, "w") as f:
        header = "#k-distance frequency[THz]\n"
        f.write(header)

        for band in bs.bands:
            for d, e in zip(bs.distance, band):
                f.write(f"{d:.8f} {e:.8f}\n")
            f.write("\n")

    return filename


def _bs_from_filename(
    filename,
    poscar,
    dim,
    symprec,
    spg,
    kpt_list,
    labels,
    primitive_axis,
    units,
    born,
    mode,
    eigenvectors,
    line_density,
):
    """Analyse input files to create band structure"""

    if ".yaml" in filename:
        logging.warning(
            f"Reading pre-computed band structure from {filename}. "
            "Be aware that many phonon-bandplot options will not "
            "be relevant."
        )
        yaml_file = filename
        phonon = None
        try:
            poscar = poscar if poscar else "POSCAR"
            poscar = Poscar.from_file(poscar)
        except OSError:
            msg = "Cannot find POSCAR file, cannot generate symmetry path."
            logging.error(f"\n {msg}")
            sys.exit()

    elif (
        "FORCE_SETS" in filename or "FORCE_CONSTANTS" in filename or ".hdf5" in filename
    ):
        try:
            poscar = poscar if poscar else "POSCAR"
            poscar = Poscar.from_file(poscar)
        except OSError:
            msg = "Cannot find POSCAR file, cannot generate symmetry path."
            logging.error(f"\n {msg}")
            sys.exit()

        if not dim:
            logging.info(
                "Supercell size (--dim option) not provided.\n"
                "Attempting to guess supercell dimensions.\n"
            )
            try:
                sposcar = Poscar.from_file("SPOSCAR")
            except OSError:
                msg = "Could not determine supercell size (use --dim flag)."
                logging.error(f"\n {msg}")
                sys.exit()

            dim = sposcar.structure.lattice.matrix @ poscar.structure.lattice.inv_matrix

            # round due to numerical noise error
            dim = np.around(dim, 5)

        elif len(dim) == 9:
            dim = np.array(dim).reshape(3, 3)

        elif np.array(dim).shape != (3, 3):
            dim = np.diagflat(dim)

        logging.info("Using supercell with dimensions:")
        logging.info("\t" + str(dim).replace("\n", "\n\t") + "\n")

        factors = {
            "ev": VaspToEv,
            "thz": VaspToTHz,
            "mev": VaspToEv * 1000,
            "cm-1": VaspToCm,
        }

        phonon = load_phonopy(
            filename,
            poscar.structure,
            dim,
            primitive_matrix=primitive_axis,
            factor=factors[units.lower()],
            symmetrise=True,
            born=born,
            write_fc=False,
        )
    elif ".phonon" in filename:
        logging.warning(
            "Reading pre-computed band structure from CASTEP. "
            "Be aware that many phonon-bandplot options will not "
            "be relevant."
        )
        castep_phonon = CastepPhonon.from_file(filename)

        cell_file = ".".join(filename.split(".")[:-1] + ["cell"])
        if isfile(cell_file):
            logging.info(f"Found .cell file, reading x-axis labels from {cell_file}")
            castep_phonon.set_labels_from_file(cell_file)
        else:
            logging.warning("No .cell file found, cannot read x-axis labels.")

        factors = {"cm-1": 1.0, "thz": 1 / THzToCm, "ev": CmToEv, "mev": 1000 * CmToEv}
        castep_phonon.frequencies *= factors[units.lower()]
        bs = castep_phonon.get_band_structure()
        return bs, None

    else:
        msg = f"Do not recognise file type of {filename}"
        logging.error(f"\n {msg}")
        sys.exit()

    # calculate band structure
    kpath, kpoints, labels = get_path_data(
        poscar.structure,
        mode=mode,
        symprec=symprec,
        spg=spg,
        kpt_list=kpt_list,
        labels=labels,
        phonopy=True,
        line_density=line_density,
    )

    # todo: calculate dos and plot also
    # phonon.set_mesh(mesh, is_gamma_center=False, is_eigenvectors=True,
    #                 is_mesh_symmetry=False)
    # phonon.set_partial_DOS()

    if ".yaml" not in filename:
        phonon.run_band_structure(
            kpoints, with_eigenvectors=eigenvectors, labels=labels
        )
        yaml_file = "sumo_band.yaml"
        phonon.band_structure.write_yaml(filename=yaml_file)

    bs = get_ph_bs_symm_line(yaml_file, has_nac=False, labels_dict=kpath.kpoints)
    return bs, phonon


def _get_parser():
    parser = argparse.ArgumentParser(
        description="""
    phonon-bandplot is a script to produce publication ready phonon band
    structure diagrams""",
        epilog=f"""
    Author: {__author__}
    Version: {__version__}
    Last updated: {__date__}""",
    )

    parser.add_argument(
        "-f",
        "--filename",
        default=None,
        metavar="F",
        help="FORCE_SETS, FORCE_CONSTANTS or band.yaml file",
    )
    parser.add_argument(
        "-p", "--prefix", metavar="P", help="prefix for the files generated."
    )
    parser.add_argument(
        "-d", "--directory", metavar="D", help="output directory for files"
    )
    parser.add_argument(
        "-q",
        "--qmesh",
        nargs=3,
        metavar="N",
        default=(8, 8, 8),
        help="q-mesh to use for phonon DOS",
    )
    parser.add_argument("-b", "--born", metavar="B", help="born effective charge file")
    parser.add_argument(
        "-e",
        "--eigenvectors",
        action="store_true",
        default=None,
        help="write the phonon eigenvectors to yaml file",
    )
    parser.add_argument(
        "--dim", nargs="+", metavar="N", help="supercell matrix dimensions"
    )
    parser.add_argument(
        "--poscar",
        default=None,
        metavar="POS",
        help="path to POSCAR file (if FORCE_SETS used)",
    )
    prim_axis = parser.add_mutually_exclusive_group()
    prim_axis.add_argument(
        "--primitive-axis",
        type=float,
        nargs=9,
        metavar="M",
        dest="primitive_axis",
        default=None,
        help="conventional to primitive cell transformation matrix",
    )
    prim_axis.add_argument(
        "--primitive-auto",
        action="store_const",
        dest="primitive_axis",
        const="auto",
        help="Let phonopy automatically determine primitive cell transformation",
    )
    parser.add_argument(
        "--symprec",
        default=0.01,
        type=float,
        help="tolerance for finding symmetry (default: 0.01)",
    )
    parser.add_argument(
        "--units",
        "-u",
        metavar="UNITS",
        default="THz",
        choices=("THz", "thz", "cm-1", "eV", "ev", "meV", "mev"),
        help=("choose units of phonon frequency (THz, cm-1, eV, meV)"),
    )
    parser.add_argument(
        "--spg", type=str, default=None, help="space group number or symbol"
    )
    parser.add_argument(
        "--density",
        type=int,
        default=60,
        help="k-point density along high-symmetry path",
    )
    parser.add_argument(
        "--seekpath",
        action="store_true",
        help="use seekpath to generate the high-symmetry path",
    )
    parser.add_argument(
        "--pymatgen",
        action="store_true",
        help="use pymatgen to generate the high-symmetry path",
    )
    parser.add_argument(
        "--cartesian", action="store_true", help="use cartesian k-point coordinates"
    )
    parser.add_argument(
        "--kpoints",
        type=str,
        default=None,
        help=('specify a list of kpoints (e.g. "0 0 0, 0.5 0 0")'),
    )
    parser.add_argument(
        "--labels",
        type=str,
        default=None,
        help=r'specify the labels for kpoints (e.g. "\Gamma,X")',
    )
    parser.add_argument(
        "--height", type=float, default=None, help="height of the graph"
    )
    parser.add_argument("--width", type=float, default=None, help="width of the graph")
    parser.add_argument(
        "--ymin", type=float, default=None, help="minimum energy on the y-axis"
    )
    parser.add_argument(
        "--ymax", type=float, default=None, help="maximum energy on the y-axis"
    )
    parser.add_argument(
        "--style",
        type=str,
        nargs="+",
        default=None,
        help="matplotlib style specifications",
    )
    parser.add_argument(
        "--no-base-style",
        action="store_true",
        dest="no_base_style",
        help="prevent use of sumo base style",
    )
    parser.add_argument(
        "--config", type=str, default=None, help="colour configuration file"
    )
    parser.add_argument(
        "--format",
        type=str,
        default="pdf",
        dest="image_format",
        metavar="FORMAT",
        help="image file format (options: pdf, svg, jpg, png)",
    )
    parser.add_argument(
        "--dpi", type=int, default=None, help="pixel density for image file"
    )
    parser.add_argument("--font", default=None, help="font to use")
    parser.add_argument(
        "--dos",
        nargs="?",
        type=str,
        default=None,
        const=True,
        help="Phonopy .dat file for phonon DOS",
    )
    parser.add_argument(
        "--to-json", type=str, default=None, help="Output JSON file for band data"
    )
    parser.add_argument(
        "--from-json",
        type=str,
        nargs="+",
        default=None,
        dest="from_json",
        help="Overlay band data from JSON files",
    )
    parser.add_argument(
        "--to-web",
        type=str,
        default=None,
        dest="to_web",
        help="(Output JSON file for http://henriquemiranda.github.io/phononwebsite)",
    )
    parser.add_argument(
        "--legend",
        type=str,
        nargs="*",
        default=None,
        help=(
            "Legend labels. With no args, label json inputs "
            "with filenames. With one arg per json file, "
            "label json inputs only. With one extra arg, "
            "label all lines beginning with new plot."
        ),
    )
    return parser


def main():
    args = _get_parser().parse_args()
    logging.basicConfig(
        filename="sumo-phonon-bandplot.log",
        level=logging.INFO,
        filemode="w",
        format="%(message)s",
    )
    console = logging.StreamHandler()
    logging.info(" ".join(sys.argv[:]))
    logging.getLogger("").addHandler(console)

    mode = "bradcrack"
    if args.seekpath:
        mode = "seekpath"
    elif args.pymatgen:
        mode = "pymatgen"

    spg = args.spg
    if args.spg:
        try:
            spg = int(spg)
        except ValueError:
            pass

    kpoints = None
    if args.kpoints:
        kpoints = [
            [list(map(float, kpt.split())) for kpt in kpts.split(",")]
            for kpts in args.kpoints.split("|")
        ]
    labels = None
    if args.labels:
        labels = [path.split(",") for path in args.labels.split("|")]

    dim = list(map(float, args.dim)) if args.dim else None

    warnings.filterwarnings("ignore", category=UserWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UnicodeWarning, module="matplotlib")
    warnings.filterwarnings("ignore", category=UserWarning, module="pymatgen")

    if args.primitive_axis == "auto":
        pa = args.primitive_axis
    elif args.primitive_axis:
        pa = np.reshape(args.primitive_axis, (3, 3))
    else:
        pa = None

    # Detect 'const' option for --legend and set filenames as legend
    legend = args.legend
    if isinstance(legend, list):
        if len(legend) == 0:
            if args.from_json:
                legend = args.from_json.copy()
            else:
                raise ValueError(
                    "Can't generate automatic legend without"
                    "--from-json; nothing to plot!"
                )

    phonon_bandplot(
        args.filename,
        poscar=args.poscar,
        prefix=args.prefix,
        directory=args.directory,
        dim=dim,
        born=args.born,
        qmesh=args.qmesh,
        primitive_axis=pa,
        symprec=args.symprec,
        units=args.units,
        spg=spg,
        line_density=args.density,
        mode=mode,
        kpt_list=kpoints,
        labels=labels,
        height=args.height,
        width=args.width,
        ymin=args.ymin,
        ymax=args.ymax,
        image_format=args.image_format,
        style=args.style,
        no_base_style=args.no_base_style,
        dpi=args.dpi,
        fonts=args.font,
        eigenvectors=args.eigenvectors,
        dos=args.dos,
        to_json=args.to_json,
        from_json=args.from_json,
        to_web=args.to_web,
        legend=legend,
    )


if __name__ == "__main__":
    main()



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
This module provides helper functions for dealing with phonopy.
"""

import logging
import sys

from phonopy import Phonopy, file_IO
from phonopy.structure.cells import determinant
from phonopy.units import Bohr, Hartree, VaspToTHz
from pymatgen.io.phonopy import get_phonopy_structure


def load_phonopy(
    filename,
    structure,
    dim,
    symprec=1e-5,
    primitive_matrix=None,
    factor=VaspToTHz,
    symmetrise=True,
    born=None,
    write_fc=False,
):
    """Load phonopy output and return an ``phonopy.Phonopy`` object.

    Args:
        filename (str): Path to phonopy output. Can be any of ``FORCE_SETS``,
            ``FORCE_CONSTANTS``, or ``force_constants.hdf5``.
        structure (:obj:`~pymatgen.core.structure.Structure`): The unitcell
            structure.
        dim (list): The supercell size, as a :obj:`list` of :obj:`float`.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.
        primitive_matrix (:obj:`list` or :obj:`str`, optional): The
            transformation matrix from the conventional to primitive cell. Only
            required when the conventional cell was used as the starting
            structure. Should be provided as a 3x3 :obj:`list` of :obj:`float`.
            Alternatively the str 'auto' may be provided.
        factor (:obj:`float`, optional): The conversion factor for phonon
            frequency. Defaults to :obj:`phonopy.units.VaspToTHz`.
        symmetrise (:obj:`bool`, optional): Symmetrise the force constants.
            Defaults to ``True``.
        born (:obj:`str`, optional): Path to file containing Born effective
            charges. Should be in the same format as the file produced by the
            ``phonopy-vasp-born`` script provided by phonopy.
        write_fc (:obj:`bool` or :obj:`str`,  optional): Write the force
            constants to disk. If ``True``, a ``FORCE_CONSTANTS`` file will be
            written. Alternatively, if set to ``"hdf5"``, a
            ``force_constants.hdf5`` file will be written. Defaults to
            ``False`` (force constants not written).
    """
    unitcell = get_phonopy_structure(structure)
    num_atom = unitcell.get_number_of_atoms()
    num_satom = determinant(dim) * num_atom

    phonon = Phonopy(
        unitcell, dim, primitive_matrix=primitive_matrix, factor=factor, symprec=symprec
    )

    if "FORCE_CONSTANTS" in filename or ".hdf5" in filename:
        # if force constants exist, use these to avoid recalculating them
        if ".hdf5" in filename:
            fc = file_IO.read_force_constants_hdf5(filename)

        elif "FORCE_CONSTANTS" in filename:
            fc = file_IO.parse_FORCE_CONSTANTS(filename=filename)

        if fc.shape[0] != num_satom:
            msg = (
                "\nNumber of atoms in supercell is not consistent with the "
                "matrix shape of\nforce constants read from {}.\nPlease"
                "carefully check --dim."
            )
            logging.error(msg.format(filename))
            sys.exit()
        phonon.force_constants = fc

    elif "FORCE_SETS" in filename:
        # load the force sets from file and calculate force constants
        fs = file_IO.parse_FORCE_SETS(filename=filename)

        if fs["natom"] != num_satom:
            msg = (
                "\nNumber of atoms in supercell is not consistent with the "
                "the data in FORCE_SETS\nPlease carefully check --dim."
            )
            logging.error(msg.format(filename))
            sys.exit()

        phonon.dataset = fs

        logging.info("Calculating force constants...")
        phonon.produce_force_constants()

    if symmetrise:
        phonon.symmetrize_force_constants()

    if born:
        # load born parameters from a file
        nac_params = file_IO.parse_BORN(
            phonon._primitive, symprec=symprec, filename=born
        )
        # set the nac unit conversion factor manual,  specific to VASP
        nac_params["factor"] = Hartree * Bohr
        phonon.nac_params = nac_params

    if write_fc == "hdf5":
        file_IO.write_force_constants_to_hdf5(phonon.force_constants)
        logging.info("Force constants written to force_constants.hdf5.")

    elif write_fc:
        file_IO.write_FORCE_CONSTANTS(phonon.force_constants)
        logging.info("Force constants written to FORCE_CONSTANTS.")

    return phonon



import errno
import logging
import math
import os
import shutil
import sys

from pymatgen.io.vasp.inputs import Kpoints


def write_kpoint_files(
    filename,
    kpoints,
    labels,
    make_folders=False,
    ibzkpt=None,
    kpts_per_split=None,
    directory=None,
    cart_coords=False,
):
    r"""Write the k-points data to files.

    Folders are named as 'split-01', 'split-02', etc ...
    KPOINTS files are named KPOINTS_band_split_01 etc ...

    Args:
        filename (:obj:`str`): Path to VASP structure file.
        kpoints (:obj:`numpy.ndarray`): The k-point coordinates along the
            high-symmetry path. For example::

                [[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0], [0.5, 0, 0.25],
                [0.5, 0, 0.5]]

        labels (:obj:`list`) The high symmetry labels for each k-point (will be
            an empty :obj:`str` if the k-point has no label). For example::

                ['\Gamma', '', 'X', '', 'Y']

        make_folders (:obj:`bool`, optional): Generate folders and copy in
            required files (INCAR, POTCAR, POSCAR, and possibly CHGCAR) from
            the current directory.
        ibzkpt (:obj:`str`, optional): Path to IBZKPT file. If set, the
            generated k-points will be appended to the k-points in this file
            and given a weight of 0. This is necessary for hybrid band
            structure calculations.
        kpts_per_split (:obj:`int`, optional): If set, the k-points are split
            into separate k-point files (or folders) each containing the number
            of k-points specified. This is useful for hybrid band structure
            calculations where it is often intractable to calculate all
            k-points in the same calculation.
        directory (:obj:`str`, optional): The output file directory.
        cart_coords (:obj:`bool`, optional): Whether the k-points are returned
            in cartesian or reciprocal coordinates. Defaults to ``False``
            (fractional coordinates).
    """
    if kpts_per_split:
        kpt_splits = [
            kpoints[i : i + kpts_per_split]
            for i in range(0, len(kpoints), kpts_per_split)
        ]
        label_splits = [
            labels[i : i + kpts_per_split]
            for i in range(0, len(labels), kpts_per_split)
        ]
    else:
        kpt_splits = [kpoints]
        label_splits = [labels]

    if cart_coords:
        coord_type = "cartesian"
        style = Kpoints.supported_modes.Cartesian
    else:
        coord_type = "reciprocal"
        style = Kpoints.supported_modes.Reciprocal

    kpt_files = []
    for kpt_split, label_split in zip(kpt_splits, label_splits):
        if ibzkpt:
            # hybrid calculation so set k-point weights to 0
            kpt_weights = ibzkpt.kpts_weights + [0] * len(kpt_split)
            kpt_split = ibzkpt.kpts + kpt_split
            label_split = [""] * len(ibzkpt.labels) + label_split
        else:
            # non-SCF calculation so set k-point weights to 1
            kpt_weights = [1] * len(kpt_split)

        segment = " -> ".join([label for label in label_split if label])
        kpt_file = Kpoints(
            comment=segment,
            num_kpts=len(kpt_split),
            kpts=kpt_split,
            kpts_weights=kpt_weights,
            style=style,
            coord_type=coord_type,
            labels=label_split,
        )
        kpt_files.append(kpt_file)

    pad = int(math.floor(math.log10(len(kpt_files)))) + 2
    if make_folders:
        for i, kpt_file in enumerate(kpt_files):
            folder = f"split-{str(i + 1).zfill(pad)}"
            if directory:
                folder = os.path.join(directory, folder)

            try:
                os.makedirs(folder)
            except OSError as e:
                if e.errno == errno.EEXIST:
                    logging.error("\nERROR: Folders already exist, won't overwrite.")
                    sys.exit()
                else:
                    raise

            kpt_file.write_file(os.path.join(folder, "KPOINTS"))
            vasp_files = [filename, "INCAR", "POTCAR", "job"]
            vasp_files += [] if ibzkpt else ["CHGCAR"]
            for vasp_file in vasp_files:
                if os.path.isfile(vasp_file):
                    shutil.copyfile(vasp_file, os.path.join(folder, vasp_file))
    else:
        for i, kpt_file in enumerate(kpt_files):
            if len(kpt_files) > 1:
                kpt_filename = f"KPOINTS_band_split_{i + 1:0d}"
            else:
                kpt_filename = "KPOINTS_band"
            if directory:
                kpt_filename = os.path.join(directory, kpt_filename)
            kpt_file.write_file(kpt_filename)



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing class for generating custom k-point paths.
"""

from itertools import chain

from sumo.symmetry import Kpath


class CustomKpath(Kpath):
    """Class to generate k-points along custom k-point paths.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.
        kpt_list (list): List of k-points to use, formatted as a list of
            subpaths, each containing a list of fractional k-points. For
            example::

                [ [[0., 0., 0.], [0., 0., 0.5]],
                  [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]

            Will return points along ``0 0 0 -> 0 0 1/2 | 1/2 0 0
            -> 1/2 1/2 0``
        path_labels (:obj:`list`): The k-point labels. These should be provided
            as a :obj:`list` of :obj:`str` for each subpath of the overall
            path. For example::

                [ ['Gamma', 'Z'], ['X', 'M'] ]

            combined with the above example for ``kpt_list`` would indicate the
            path: Gamma -> Z | X -> M. If no labels are provided, letters from
            A -> Z will be used instead.

    Attributes:
        prim (:obj:`~pymatgen.core.structure.Structure`): The standardised
            primitive cell structure for the generated k-point path.
        conv (:obj:`~pymatgen.core.structure.Structure`): The standardised
            conventional cell structure.
    """

    def __init__(self, structure, kpt_list, path_labels=None, symprec=1e-3):
        Kpath.__init__(self, structure, symprec=symprec)

        if path_labels is None:
            path_labels = self._auto_kpath_labels(kpt_list)

        try:
            assert len(kpt_list) == len(path_labels)
            for kpt_segment, path_segment in zip(kpt_list, path_labels):
                assert len(kpt_segment) == len(path_segment)
        except AssertionError:
            raise ValueError("kpt_list and path_labels are not consistent")

        flat_kpts = chain(*kpt_list)
        flat_path_labels = chain(*path_labels)
        kpt_dict = {label: kpt for label, kpt in zip(flat_path_labels, flat_kpts)}

        self._kpath = {"kpoints": kpt_dict, "path": path_labels}

    @staticmethod
    def _auto_kpath_labels(kpt_list):
        """Get a default set of labels (1), (2), (3)... for a k-point path

        Repeated points will be identified and the labels re-used.

        Args:
            kpt_list (list): Nested list representing k-point path segments,
                e.g.::

                  [[[0., 0., 0.], [0., 0., 0.5], [0., 0.5, 0.5]],
                   [[0.5, 0.5, 0.], [0., 0., 0.]]]

        Returns:
            list: Corresponding nested list of labels, e.g.::

              [['(1)', '(2)', '(3)'], ['(4)', '(1)']]

        """

        # Build dict of labels
        label_i = 1
        kpt_labels = {}
        for kpt in chain(*kpt_list):
            if tuple(kpt) in kpt_labels:
                continue
            else:
                kpt_labels.update({tuple(kpt): f"({label_i})"})
                label_i += 1

        # Read out into nested lists
        kpath_labels = [
            [kpt_labels[tuple(kpt)] for kpt in segment] for segment in kpt_list
        ]

        return kpath_labels



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing class for generating Bradley and Cracknell k-point paths.
"""

from json import load as load_json

import numpy as np
import pkg_resources

from sumo.symmetry import Kpath


class BradCrackKpath(Kpath):
    r"""Class to generate k-points along paths from Bradley and Cracknell.

    The paths used are based on Brillouin zones depicted in reference [brad]_.

    .. [brad] "The Mathematical Theory of Symmetry in Solids", C. J. Bradley
              and A. P. Cracknell, Clarendon Press, 1972.

    These paths represent only one particular route through the Brillouin
    zone and do not cover every possible path (though they do visit every
    high-symmetry k-point at least once).

    These paths should be used with the correct primitive structure. This
    structure can be accessed using the ``prim`` attribute. Compliance
    between the provided structure and standardised structure checked using the
    ``correct_structure()`` method.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.
        spg (:obj:`~pymatgen.symmetry.groups.SpaceGroup`, optional): Space
            group used to override the symmetry determined by spglib. This is
            not recommended and only provided for testing purposes.

    Attributes:
        prim (:obj:`~pymatgen.core.structure.Structure`): The standardised
            primitive cell structure for the generated k-point path.
        conv (:obj:`~pymatgen.core.structure.Structure`): The standardised
            conventional cell structure.
    """

    def __init__(self, structure, symprec=1e-3, spg=None):
        Kpath.__init__(self, structure, symprec=symprec)

        angles = self.conv.lattice.angles
        unique = angles.index(min(angles, key=angles.count))
        a = self.conv.lattice.abc[0]
        b = self.conv.lattice.abc[1]
        c = self.conv.lattice.abc[2]

        if spg:
            spg_symbol = spg.symbol
            lattice_type = self.get_lattice_type(spg.int_number)
        else:
            spg_symbol = self.spg_symbol
            lattice_type = self.lattice_type

        bravais = self._get_bravais_lattice(spg_symbol, lattice_type, a, b, c, unique)
        self._kpath = self._get_bradcrack_data(bravais)

    @staticmethod
    def _get_bradcrack_data(bravais):
        r"""Read Bradley--Cracknell k-points path from data file

        Args:
            bravais (str): Lattice code including orientation e.g. 'trig_p_c'

        Returns:
            dict: kpoint path and special point locations, formatted as e.g.::

              {'kpoints': {'\Gamma': [0., 0., 0.], 'X': [0., 0.5, 0.], ...},
               'path': [['\Gamma', 'X', ..., 'P'], ['H', 'N', ...]]}

        """
        json_file = pkg_resources.resource_filename(__name__, "bradcrack.json")
        with open(json_file) as f:
            bradcrack_data = load_json(f)
            return bradcrack_data[bravais]

    @staticmethod
    def _get_bravais_lattice(spg_symbol, lattice_type, a, b, c, unique):
        """Get Bravais lattice symbol from symmetry data"""

        if lattice_type == "triclinic":
            return "triclinic"

        elif lattice_type == "monoclinic":
            if "P" in spg_symbol:
                if unique == 0:
                    return "mon_p_a"
                elif unique == 1:
                    return "mon_p_b"
                elif unique == 2:
                    return "mon_p_c"

            elif "C" in spg_symbol:
                if unique == 0:
                    return "mon_c_a"
                elif unique == 1:
                    return "mon_c_b"
                elif unique == 2:
                    return "mon_c_c"

        elif lattice_type == "orthorhombic":
            if "P" in spg_symbol:
                return "orth_p"

            elif "A" in spg_symbol or "C" in spg_symbol:
                if a > b:
                    return "orth_c_a"
                elif b > a:
                    return "orth_c_b"

            elif "F" in spg_symbol:
                if (
                    1 / a**2 < 1 / b**2 + 1 / c**2
                    and 1 / b**2 < 1 / c**2 + 1 / a**2
                    and 1 / c**2 < 1 / a**2 + 1 / b**2
                ):
                    return "orth_f_1"
                elif 1 / c**2 > 1 / a**2 + 1 / b**2:
                    return "orth_f_2"
                elif 1 / b**2 > 1 / a**2 + 1 / c**2:
                    return "orth_f_3"
                elif 1 / a**2 > 1 / c**2 + 1 / b**2:
                    return "orth_f_4"

            elif "I" in spg_symbol:
                if a > b and a > c:
                    return "orth_i_a"
                elif b > a and b > c:
                    return "orth_i_b"
                elif c > a and c > b:
                    return "orth_i_c"

        elif lattice_type == "tetragonal":
            if "P" in spg_symbol:
                return "tet_p"

            elif "I" in spg_symbol:
                if a > c:
                    return "tet_i_a"
                else:
                    return "tet_i_c"

        elif (
            lattice_type == "trigonal"
            or lattice_type == "hexagonal"
            or lattice_type == "rhombohedral"
        ):
            if "R" in spg_symbol:
                if a > np.sqrt(2) * c:
                    return "trig_r_a"
                else:
                    return "trig_r_c"

            elif "P" in spg_symbol:
                if unique == 0:
                    return "trig_p_a"
                elif unique == 2:
                    return "trig_p_c"

        elif lattice_type == "cubic":
            if "P" in spg_symbol:
                return "cubic_p"
            elif "I" in spg_symbol:
                return "cubic_i"
            elif "F" in spg_symbol:
                return "cubic_f"



import errno
import logging
import os.path
import re
from itertools import chain, product
from math import ceil
from os import makedirs
from shutil import which
from subprocess import PIPE, Popen

import numpy as np
from monty.io import zopen
from pymatgen.core.lattice import Lattice
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos

import sumo.electronic_structure.dos
from sumo.electronic_structure.optics import kkr

_bohr_to_angstrom = 0.5291772
_ry_to_ev = 13.605693009


class QuestaalSite:
    """Structure information: Questaal site.ext file

    Usually this will be instantiated with the
    :obj:`~sumo.io.questaal.QuestaalSite.from_file()` method.

    """

    def __init__(
        self,
        nbas,
        vn=3.0,
        io=15,
        alat=1.0,
        xpos=True,
        read="fast",
        sites=None,
        plat=(1, 0, 0, 0, 1, 0, 0, 0, 1),
    ):
        sites = sites or []
        if nbas != len(sites):
            raise AssertionError()
        if len(plat) != 9:
            raise AssertionError()

        if read != "fast":
            raise Exception("Algebraic expressions not supported, use 'fast'")
        if io != 15:
            raise Exception("Only site.ext format 15 supported at present")

        self.nbas, self.vn, self.io, self.alat = nbas, vn, io, alat
        self.xpos, self.read, self.sites, self.plat = xpos, read, sites, plat

        is_empty = re.compile(r"E\d*$")
        empty_sites = [
            site for site in sites if is_empty.match(site["species"]) is not None
        ]
        self.nbas_empty = len(empty_sites)

    @property
    def structure(self):
        # Get lattice vectors in Angstrom
        lattice = Lattice(self.plat)
        lattice = Lattice(lattice.matrix * self.alat * _bohr_to_angstrom)

        # Get corresponding lists of species and positions by making a list of
        # pairs and unpacking with zip
        if self.xpos:
            species_coords = [(site["species"], site["pos"]) for site in self.sites]
            species, coords = zip(*species_coords)

            return Structure(lattice, species, coords, coords_are_cartesian=False)
        else:
            species_coords = [
                (
                    site["species"],
                    [x * self.alat * _bohr_to_angstrom for x in site["pos"]],
                )
                for site in self.sites
            ]
            species, coords = zip(*species_coords)

            return Structure(lattice, species, coords, coords_are_cartesian=True)

    @classmethod
    def from_file(cls, filename):
        with zopen(filename, "rt") as f:
            lines = f.readlines()

        header = lines[0]
        sites = [line for line in lines if line[0] not in "#%"]

        # Some of the header info does not use '=' so handle separately
        header_items = header.strip().split()
        if header_items[0] != "%" or header_items[1] != "site-data":
            raise AssertionError()

        xpos = True if "xpos" in header_items else False
        read = "fast" if "fast" in header_items else False

        header_clean = " ".join(
            x for x in header_items if x not in ("%", "site-data", "xpos", "fast")
        )

        tags = re.findall(r"(\w+)\s*=", header_clean)  # Find tags
        # Split on tags to find tag parameters
        tag_data = re.split(r"\s*\w+\s*=\s*", header_clean)[1:]
        tag_dict = dict(zip(tags, tag_data))

        vn = float(tag_dict["vn"]) if "vn" in tag_dict else 3.0
        io = int(tag_dict["io"]) if "io" in tag_dict else 15.0
        nbas = int(tag_dict["nbas"]) if "nbas" in tag_dict else 15.0
        alat = float(tag_dict["alat"]) if "alat" in tag_dict else 1.0
        plat = (
            [float(x) for x in tag_dict["plat"].split()]
            if "plat" in tag_dict
            else [1, 0, 0, 0, 1, 0, 0, 0, 1]
        )

        # Convert sites to structured format
        # (If needed, could support other 'io' options here)
        sites = [
            {"species": site.split()[0], "pos": [float(x) for x in site.split()[1:4]]}
            for site in sites
        ]

        return cls(
            nbas, vn=vn, io=io, alat=alat, xpos=xpos, read=read, sites=sites, plat=plat
        )


class QuestaalInit:
    """Structure information: Questaal init.ext file

    Usually this will be instantiated with the
    :obj:`~sumo.io.questaal.QuestaalInit.from_file()` method.
    Data from each file section is given as a separate input argument and
    stored as a property.

    Args:
        lattice (:obj:`dict`):
            There are two main forms of lattice data (stored in self.lattice):
            - Explicit lattice vectors expressed with PLAT
            - Lattice angles/lengths expressed as A, B, C, ALPHA, BETA, GAMMA
              - This style of input must be used with SPCGRP, specifying
                symmetry
              - Angles are given in degrees

            Examples:

                lattice = {'SPCGRP': 186, 'A': 3.18409958, 'C': 5.1551,
                           'UNITS': 'A', 'ALAT': 1}

                lattice = {'ALAT': 1, 'UNITS': 'A',
                           'PLAT': [[1.59, -2.75, 0.],
                                    [1.59, 2.75, 0.],
                                    [0., 0., 5.16]]}
        site (:obj:`list`): Site species, coordinate type and location as a
            list of dictionaries of form
            ``{'ATOM': el, COORD_TYPE: (a, b, c)}``
            where *el* is the species label, COORD_TYPE is ``'POS'``
            (Cartesian) or ``'X'`` (direct/fractional) and (a, b, c) is a tuple
            giving the site position in Cartesian or fractional coordinates.

        spec (:obj:`dict`, optional): Content of SPEC section, expressed as
            Python dict

        tol (:obj:`float`, optional): Tolerance in units of distance used when
            constructing cell from space group

        ignore_units (:obj:`bool`, optional): If True, no unit conversions will
            be applied when converting to other formats (e.g. Pymatgen
            structure). This is needed for consistent use of Bohr units when
            interpreting band structure data.

    """

    def __init__(self, lattice, site, spec=None, tol=1e-5, ignore_units=False):
        self.lattice = lattice
        self.site = site
        self.spec = spec
        self.tol = tol
        self.ignore_units = ignore_units

        cartesian_sites = any("POS" in item for item in site)
        fractional_sites = any("X" in item for item in site)
        c_sites = any("C" in item for item in site)

        if c_sites:
            raise NotImplementedError(
                "C position option for Questaal input "
                "(conventional lattice vector fractions)"
                " not implemented. Life is too short!"
            )
        if cartesian_sites and fractional_sites:
            raise ValueError("Cannot mix direct and Cartesian input")
        else:
            self.cartesian = cartesian_sites

    @property
    def structure(self):
        """Pymatgen structure object from Questaal init file"""

        if "SPCGRP" in self.lattice and self.lattice["SPCGRP"]:
            return self._get_structure_from_spcgrp()
        else:
            return self._get_structure_from_lattice()

    def _get_species_coords(self):
        species = [entry["ATOM"] for entry in self.site]

        if self.cartesian:
            coords = [entry["POS"] for entry in self.site]
        else:
            coords = [entry["X"] for entry in self.site]

        return species, coords

    def _get_structure_from_spcgrp(self):
        if "A" not in self.lattice:
            raise AssertionError()
        if "B" not in self.lattice:
            logging.info("Lattice vector B not given, assume equal to A")
            self.lattice["B"] = self.lattice["A"]
        if "C" not in self.lattice:
            logging.info("Lattice vector C not given, assume equal to A")
            self.lattice["C"] = self.lattice["C"]
        if "ALPHA" not in self.lattice:
            logging.info("Lattice angle ALPHA not given, assume right-angle")
            self.lattice["ALPHA"] = 90
        if "BETA" not in self.lattice:
            logging.info("Lattice angle BETA not given, assume right-angle")
            self.lattice["BETA"] = 90
        if "GAMMA" not in self.lattice:
            try:
                spcgrp_number = int(self.lattice["SPCGRP"])
            except ValueError:
                spcgrp_number = 0
            if 167 < spcgrp_number < 195:
                logging.info(
                    "Lattice angle GAMMA not given, "
                    "hexagonal space group, assume 120"
                )
                self.lattice["GAMMA"] = 120
            else:
                logging.info("Lattice angle GAMMA not given, assume right-angle")
                self.lattice["GAMMA"] = 90

        if self.cartesian:
            logging.info(
                "Warning: Cartesian positions used without explicit lattice vectors"
            )

        if "UNITS" not in self.lattice or self.lattice["UNITS"] is None:
            if self.ignore_units:
                pass
            else:
                for length in ("A", "B", "C"):
                    self.lattice[length] *= _bohr_to_angstrom

        if "ALAT" not in self.lattice:
            raise AssertionError()
        for length in ("A", "B", "C"):
            self.lattice[length] *= self.lattice["ALAT"]

        lattice = Lattice.from_parameters(
            self.lattice["A"],
            self.lattice["B"],
            self.lattice["C"],
            self.lattice["ALPHA"],
            self.lattice["BETA"],
            self.lattice["GAMMA"],
        )

        species, coords = self._get_species_coords()

        return Structure.from_spacegroup(
            self.lattice["SPCGRP"],
            lattice,
            species,
            coords,
            coords_are_cartesian=self.cartesian,
            tol=self.tol,
        )

    def _get_structure_from_lattice(self):
        lattice = Lattice(self.lattice["PLAT"])
        lattice = Lattice(lattice.matrix * self.lattice["ALAT"])
        species, coords = self._get_species_coords()

        if self.ignore_units:
            pass
        else:
            if "UNITS" not in self.lattice or self.lattice["UNITS"] is None:
                lattice = Lattice(lattice.matrix * _bohr_to_angstrom)
                coords = [
                    tuple(x * _bohr_to_angstrom for x in site_coords)
                    for site_coords in coords
                ]

        return Structure(lattice, species, coords, coords_are_cartesian=self.cartesian)

    def to_file(self, filename):
        """Write QuestaalInit object to init file"""
        with open(filename, "w") as f:

            f.write("LATTICE\n")
            for key, value in self.lattice.items():
                if key == "PLAT":
                    #  Expand nested lists to one flat list.
                    #  Yes, nested list comprehensions look weird!
                    lattice_params = [c for row in self.lattice["PLAT"] for c in row]
                    # Write out as string-separated row of 9
                    f.write("    PLAT= " + " ".join(map(str, lattice_params)))
                    f.write("\n")
                else:
                    f.write(f"    {key}={value}\n")

            f.write("SITE\n")
            for row in self.site:
                f.write(f"    ATOM={row['ATOM']:4s}  ")
                if "POS" in self.site:
                    f.write("POS= {:11.8f} {:11.8f} {:11.8f}".format(*row["POS"]))
                else:
                    f.write("X= {:11.8f} {:11.8f} {:11.8f}".format(*row["X"]))
                for key, value in row.items():
                    if key not in ("ATOM", "POS", "X"):
                        f.write(f"  {key}= {value}")
                f.write("\n")

            if self.spec is not None:
                f.write("SPEC")
                for key, value in self.spec.items():
                    f.write(f"  {key}= {value}\n")

    @staticmethod
    def from_structure(structure):
        """Generate QuestaalInit object from pymatgen structure"""
        lattice = {"ALAT": 1, "UNITS": "A", "PLAT": structure.lattice.matrix}
        sites = [
            {"ATOM": site.species_string, "X": tuple(site.frac_coords)}
            for site in structure.sites
        ]
        return QuestaalInit(lattice, sites)

    @staticmethod
    def from_file(filename, preprocessor=None, tol=1e-5, ignore_units=False):
        """Read QuestaalInit object from init.ext file

        Args:
            filename (:obj:`str`): Path to init.ext file
            preprocessor (:obj:`bool`): Process file with ``rdfile`` (must be
                available on shell PATH). If None, use preprocessor where
                available. If True, an error will be raised if ``rdfile`` is
                unavailable.
            tol (:obj:`float`, optional): tolerance for symmetry operations
            ignore_units (:obj:`bool`, optional): If True, no unit conversions
                will be applied when converting to other formats (e.g. Pymatgen
                structure). This is needed for consistent use of Bohr units
                when interpreting band structure data.


        Returns:
            :obj:`~sumo.io.questaal.QuestaalInit`"""

        if preprocessor is None:
            if which("rdfile") is not None:
                preprocessor = True
            else:
                preprocessor = False

        if preprocessor:
            process = Popen(["rdfile", filename], stdout=PIPE)
            lines = process.stdout.readlines()
            #  Need to decode from bytes. Hard-coding ASCII here - it doesn't
            #  seem likely that Questaal would support unicode?
            lines = [line.decode("ascii") for line in lines]
        else:
            with zopen(filename, "r") as f:
                lines = f.readlines()

        categories = {"LATTICE", "SITE", "SPEC"}

        # Find which lines begin a new category
        cat_lines = []
        for i, line in enumerate(lines):
            if line.strip().split()[0] in categories:
                cat_lines.append(i)
        cat_lines.append(None)  # None allows us to slice up to the file end

        # Grab the lines corresponding to each section and collect by category
        grouped_lines = {}
        for i in range(len(cat_lines) - 1):
            category = lines[cat_lines[i]].split()[0]
            grouped_lines[category] = lines[cat_lines[i] : cat_lines[i + 1]]

        # Initial cleanup: - Remove leading/trailing whitespace
        #                  - drop lines beginning with '#'
        #                  - remove category name from first line

        for category, lines in grouped_lines.items():
            lines = [line.strip() for line in lines if line.strip()[0] != "#"]

            category_line_remainder = lines[0][len(category) :].strip()
            lines = [category_line_remainder] + lines[1:]

            grouped_lines[category] = lines

        # Join lines and split into tags
        init_data = {}
        for category, lines in grouped_lines.items():
            tag_text = " ".join(lines)

            if category == "SITE":
                site_data = []
                #  Split on regex: ATOM tag will be removed, species is left
                #  followed by other tags, e.g.
                #  "ATOM=Zn X = 0.0 0.0 0.5 ATOM = S  X= 0.0 0.0 0.0"
                #  is split to
                #  ['', 'Zn X = 0.0 0.0 0.5', 'S X = 0.0 0.0 0.0']
                #
                atom_entries = re.split(r"ATOM\s*=\s*", tag_text)

                for line in atom_entries[1:]:  # Drop the empty first line
                    atom = line.split()[0]
                    tag_dict = {"ATOM": atom}

                    line = line[len(atom) :]  # Drop species tag from line
                    tags = re.findall(r"(\w+)\s*=", line)  # Find tags
                    # Split on tags to find tag parameters
                    tag_data = re.split(r"\s*\w+\s*=\s*", line)[1:]
                    tag_dict.update(dict(zip(tags, tag_data)))

                    # Cast coordinates to tuple
                    for key in ("POS", "X"):
                        if key in tag_dict:
                            tag_dict[key] = tuple(map(float, tag_dict[key].split()))

                    site_data.append(tag_dict)

                init_data["SITE"] = site_data

            else:
                float_params = ("A", "B", "C", "ALPHA", "BETA", "GAMMA", "ALAT")

                unsupported_params = "GENS"

                tags = re.findall(r"(\w+)\s*=", tag_text)  # Find tags
                # Split on tags to find tag parameters
                tag_data = re.split(r"\s*\w+\s*=\s*", tag_text)[1:]
                tag_dict = dict(zip(tags, tag_data))

                if "SPCGRP" in tag_dict:
                    try:
                        tag_dict["SPCGRP"] = int(tag_dict["SPCGRP"])
                    except ValueError:
                        pass

                if "PLAT" in tag_dict:
                    lattice = tuple(map(float, tag_dict["PLAT"].split()))
                    if len(lattice) != 9:
                        raise AssertionError()
                    tag_dict["PLAT"] = [
                        [lattice[0], lattice[1], lattice[2]],
                        [lattice[3], lattice[4], lattice[5]],
                        [lattice[6], lattice[7], lattice[8]],
                    ]

                for float_param in float_params:
                    if float_param in tag_dict:
                        tag_dict[float_param] = float(tag_dict[float_param])

                for unsupported_param in unsupported_params:
                    if unsupported_param in tag_dict:
                        raise NotImplementedError(
                            "Questaal tag {}_{} is not supported".format(
                                category, unsupported_param
                            )
                        )

                init_data[category] = tag_dict

            if "SPEC" not in init_data or init_data["SPEC"] == {}:
                init_data["SPEC"] = None

        return QuestaalInit(
            init_data["LATTICE"], init_data["SITE"], spec=init_data["SPEC"], tol=tol
        )


def write_kpoint_files(
    filename,
    kpoints,
    labels,
    make_folders=False,
    directory=None,
    cart_coords=False,
    **kwargs,
):
    r"""Write syml file for Questaal kpoints

    The interface imitates the VASP KPOINTS file writer for simplicity of
    integration into Sumo, but there are some conceptual differences:

    If *labels* is None, then *kpoints* will be written to a simple file in
    "List mode".

    If labels are provided, then labelled points will be extracted from the
    list of kpoints and "line mode" is used to create a compact band-structure
    input file with the same number of k-points.

    Args:
        filename (:obj:`str`): Path to init.ext file. (Extension is used to
            name syml file).
        kpoints (:obj:`numpy.ndarray`): The k-point coordinates along the
            high-symmetry path. For example::

                [[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0], [0.5, 0, 0.25],
                [0.5, 0, 0.5]]

        labels (:obj:`list`) The high symmetry labels for each k-point (will be
            an empty :obj:`str` if the k-point has no label). For example::

                ['\Gamma', '', 'X', '', 'Y']

        make_folders (:obj:`bool`, optional): Generate folders and copy in
            required files if found from the current directory.
        directory (:obj:`str`, optional): The output file directory.
        cart_coords (:obj:`bool`, optional): Whether the k-points are returned
            in Cartesian or reciprocal coordinates. Defaults to ``False``
            (fractional coordinates). Note that Questaal uses crystallographic-
            style Cartesian coordinates and a factor of 1/(2 pi) should be
            included in the kpoints data.
    """

    for key, value in kwargs.items():
        if value is not None:
            logging.info(
                'Ignoring k-point write option "{}"; not '
                "implemented for Questaal calculations.".format(key)
            )

    ext = filename.split(".")[-1]
    logging.info(f"System id from init filename: {ext}")

    if directory is not None:
        path = directory
    else:
        path = os.path.curdir

    if make_folders:
        path = os.path.join(path, "band-calc")

        try:
            makedirs(path)
        except OSError as e:
            if e.errno == errno.EEXIST:
                import sys

                logging.error("\nERROR: Folders already exist, won't overwrite.")
                sys.exit()
            else:
                raise

    if cart_coords:
        logging.info(
            "Writing band structure in Cartesian coordinates...\n"
            "Remember to run full-potential calc with --band and "
            "NOT --band~mq"
        )
    else:
        logging.info(
            "Writing band structure in direct coordinates...\n"
            "Remember to run full-potential calc with --band~mq."
        )
    if labels is None:
        with open(os.path.join(path, "syml." + ext), "w") as f:
            for kpt in kpoints:
                f.write(f"{kpt[0]:11.8f} {kpt[1]:11.8f} {kpt[2]:11.8f}\n")
    else:
        label_positions = [i for i, label in enumerate(labels) if label != ""]
        special_points = [kpoints[i] for i in label_positions]
        segment_samples = [
            label_positions[i + 1] - label_positions[i] + 1
            for i in range(len(label_positions) - 1)
        ]
        with open(os.path.join(path, "syml." + ext), "w") as f:
            for i, samples in enumerate(segment_samples):
                if samples == 2:
                    continue  # Don't add segments between branches
                f.write(
                    "{:5d}    {:11.8f} {:11.8f} {:11.8f}    "
                    "{:11.8f} {:11.8f} {:11.8f}    {} to {}\n".format(
                        samples,
                        special_points[i][0],
                        special_points[i][1],
                        special_points[i][2],
                        special_points[i + 1][0],
                        special_points[i + 1][1],
                        special_points[i + 1][2],
                        labels[label_positions[i]],
                        labels[label_positions[i + 1]],
                    )
                )
            f.write("    0 0 0 0 0 0 0\n")


def labels_from_syml(syml_file):
    """Read special point locations from Questaal syml file

    Each line of file should have format:

        NPTS  X1 Y1 Z1  X2 Y2 Z2  label1 to label2

    except the last line which is seven 0 characters. Whitespace amounts
    are flexible. If a label is defined more than once, the last definition
    will be used.

    Args:
        syml_file (:obj:`str`): Path to questaal syml.ext input file. The
            locations and labels of special points are read from this file.

    Returns:
        {:obj:`str`: :obj:`tuple`, ...}

    """
    labels = {}

    with open(syml_file) as f:
        lines = f.readlines()
    for line in lines:
        npts, x1, y1, z1, x2, y2, z2, *label_text = line.split()
        if len(label_text) < 3:
            pass
        else:
            kpt1 = tuple(map(float, (x1, y1, z1)))
            kpt2 = tuple(map(float, (x2, y2, z2)))

            label_text = " ".join(label_text)  # Undo previous split
            label1_label2 = label_text.split(" to ")
            if len(label1_label2) != 2:
                raise ValueError(
                    f"Not clear how to interpret labels from this line: {line}"
                )
            label1, label2 = label1_label2
            labels.update({label1: kpt1, label2: kpt2})
    return labels


def read_dos(
    pdos_file=None,
    tdos_file=None,
    site_file=None,
    ry=True,
    total_only=False,
    gaussian=None,
    elements=None,
    lm_orbitals=None,
    atoms=None,
):
    """Read DOS file data

    Construct Pymatgen Dos objects for total dos and orbital contributions as
    available from Questaal dos.ext file

    Args:
        pdos_file (:obj:`str`):
            Path to DOS file from Questaal, as generated by LMF with ``--dos``
            option (to obtain total DOS) or with LMDOS from moms.ext (to obtain
            PDOS). It is assumed in either case that the default mode=2 was
            used (i.e. DOS is resolved by site, l and m). If there is only one
            channel of data, this is interpreted as a total DOS.
        tdos_file (:obj:`str`):
            Path to TDOS file from Questaal, as generated by LMF with ``--dos``
            option BEFORE LMDOS is used to generate a PDOS from *moms.ext*.
        site_file (:obj:`str`, optional):
            Path to site.ext file from Questaal. This is used to identify the
            sites and obtain structure information for PDOS manipulation.
        ry (:obj:`bool`, optional):
            Convert energy values from Ry to eV
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the site.ext. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.
        gaussian (:obj:`float`, optional): Broaden the density of states using
            convolution with a gaussian function. This parameter controls the
            sigma or standard deviation of the gaussian distribution.
        total_only (:obj:`bool`, optional): Only extract the total density of
            states. Defaults to ``False``.

    Returns:
        (:obj:`tuple`):
            *dos*, *pdos*: *dos* is a
            :obj:`~pymatgen.electronic_structure.dos.Dos` object containing the
            total density of states and ``pdos`` is a :obj:`dict` of
            :obj:`dict` mapping the elements and their orbitals to
            :obj:`~pymatgen.electronic_structure.dos.Dos` objects. For
            example::

                 {
                     'Bi': {'s': Dos, 'p': Dos ... },
                     'S': {'s': Dos}
                 }
    """

    def _read_states(f, nlines):
        lines = [f.readline() for i in range(nlines)]
        # This statement is very "functional programming"; read it
        # backwards.  List of split lines is "flattened" by chain into
        # iterator of values; this is fed into map to make floats and
        # stored to a list
        return list(map(float, chain(*(line.split() for line in lines))))

    def _read_dos_data(filename, check_tdos=False):
        with zopen(filename, "rt") as f:
            (emin, emax, ne, nchan, nsp, efermi, delta, fmt) = f.readline().split()

            if check_tdos and int(nchan) != 1:
                raise ValueError(
                    f"File {filename} is not a TDOS: contains {nchan} channels"
                )

            ne, nchan, nsp, fmt = map(int, (ne, nchan, nsp, fmt))
            emin, emax, efermi, delta = map(float, (emin, emax, efermi, delta))
            if ry:
                emin, emax, efermi, delta = (
                    x * _ry_to_ev for x in (emin, emax, efermi, delta)
                )

            nlines = ceil(ne / 5)
            data = [_read_states(f, nlines) for _ in range(nsp * nchan)]

        energies = np.linspace(emin, emax, ne)

        return energies, data, efermi, nsp

    def _get_tdos(filename):
        energies, data, efermi, nsp = _read_dos_data(tdos_file, check_tdos=True)

        if nsp == 1:
            # densities = np.array(data[0]).reshape((ne, 1))
            densities = {Spin.up: data[0]}
        elif nsp == 2:
            # densities = np.concatenate((data[0], data[1]), axis=1)
            densities = {Spin.up: data[0], Spin.down: data[1]}
        else:
            raise ValueError(
                f"There can't be {nsp} spin channels, that makes no sense!"
            )
        return Dos(efermi, energies, densities)

    def _get_cdos(pdos_file, site_file):
        if not (pdos_file and site_file):
            raise ValueError("Both dos.ext and site.ext are needed for PDOS")
        energies, data, efermi, nsp = _read_dos_data(pdos_file)
        if nsp == 1:
            fake_tdos = Dos(efermi, energies, {Spin.up: 0 * energies})
            spins = (Spin.up,)
        elif nsp == 2:
            fake_tdos = Dos(
                efermi, energies, {Spin.up: 0 * energies, Spin.down: 0 * energies}
            )
            spins = (Spin.up, Spin.down)
        else:
            raise ValueError(
                f"There can't be {nsp} spin channels, that makes no sense!"
            )

        site_data = QuestaalSite.from_file(site_file)
        structure = site_data.structure

        pdoss = {}
        for site, orbital, spin in product(
            range(len(site_data.sites)),
            range(16),  # forget about g orbs
            range(len(spins)),
        ):
            if structure.sites[site] not in pdoss:
                pdoss.update({structure.sites[site]: {}})
            if Orbital(orbital) not in pdoss[structure.sites[site]]:
                pdoss[structure.sites[site]].update({Orbital(orbital): {}})

            pdoss[structure.sites[site]][Orbital(orbital)].update(
                {spins[spin]: data[site * (25 * nsp) + orbital * nsp + spin]}
            )

        return CompleteDos(structure, fake_tdos, pdoss)

    if not (tdos_file or pdos_file):
        raise ValueError("Need some DOS data: provide a dos.ext")

    elif tdos_file and not pdos_file:
        tdos = _get_tdos(tdos_file)
        cdos = None

    elif pdos_file and not tdos_file:
        with zopen(pdos_file, "rt") as f:
            nchan = int(f.readline().split()[3])
        if nchan == 1:
            logging.info(
                "No explicit TDOS and file {} contains 1 channel: "
                "this is a TDOS.".format(pdos_file)
            )

            tdos = _get_tdos(pdos_file)
            cdos = None
        else:
            tdos = None
            cdos = _get_cdos(pdos_file, site_file)

    else:
        tdos = _get_tdos(tdos_file)
        cdos = _get_cdos(pdos_file, site_file)

    if gaussian:
        if tdos is not None:
            tdos.densities = tdos.get_smeared_densities(gaussian)
        if cdos is not None:
            for site in cdos.pdos:
                for orbital in cdos.pdos[site]:
                    cdos.pdos[site][orbital] = cdos.get_site_orbital_dos(
                        site, orbital
                    ).get_smeared_densities(gaussian)

    # Combine data into selected orbitals: default is by species type
    if cdos is None:
        pdos = {}
    else:
        pdos = sumo.electronic_structure.dos.get_pdos(
            cdos, lm_orbitals=lm_orbitals, atoms=atoms, elements=elements
        )

    return tdos, pdos


def band_structure(bnds_file, lattice, labels=None, alat=1, coords_are_cartesian=False):
    """Read band structure data

    Args:
        bnds_file (:obj:`str`): Path to questaal bnds.ext output file. The
            k-point positions and eigenvalues are read from this file.
        labels (:obj:`dict`): Dict of special point locations and labels. This
            is generally obtained from a syml.ext file using
            :obj:`sumo.io.questaal.labels_from_syml`.
        alat (:obj:`float`): Lattice scaling parameter defined in site.ext
        lattice (:obj:`pymatgen.core.lattice.Lattice`)
        coords_are_cartesian (:obj:`bool`): bnds.ext file is in Cartesian
            coordinates.

    Returns:
        :obj:`pymatgen.electronic_structure.bandstructure.BandStructureSymmLine`
    """

    # Implementation note: pymatgen band structure expects an eigenvalue array
    # ordered [band][kpt] but the data will be read in one kpt at a time as
    # this is the format of the bnd data file.
    #
    # We will first build a nested list [[bnd1_kpt1, bnd2_kpt1, ...],
    #                                    [bnd1_kpt2, bnd2_kpt2, ...], ...]
    # then convert to a numpy array (i.e. with each row for a kpoint)
    # and then transpose the array to obtain desired formats

    with zopen(bnds_file, "r") as f:
        kpoints = []

        # Read heading, get metadata and check no orbital projections used
        nbands, efermi, n_color_wts, *_ = f.readline().split()

        if int(n_color_wts) > 0:
            raise NotImplementedError(
                "Band data includes orbital data: "
                "this format is not currently supported."
            )

        nbands, efermi = int(nbands), float(efermi)
        eig_lines = ceil(nbands / 10)

        # Check if first two kpts are the same: if so, assume there are two
        # spin channels
        _ = f.readline()
        kpt1 = list(map(float, f.readline().split()))

        for line in range(eig_lines):  # Skip over the eigenvalues
            _ = f.readline()  # for now: re-read file later
        kpt2 = list(map(float, f.readline().split()))
        if len(kpt1) != 3 or len(kpt2) != 3:
            raise AssertionError()

        if kpt1 == kpt2:
            spin_pol = True
        else:
            spin_pol = False

    logging.info("Reading Questaal band structure header...")
    logging.info(
        "nbands: {}, efermi / Ry: {}, spin channels: {}".format(
            nbands, efermi, (2 if spin_pol else 1)
        )
    )
    logging.info("Reading band structure eigenvalues...")

    # Key info has been collected: re-read file for kpts and eigenvalues
    def _read_eigenvals(f, nlines):
        lines = [f.readline() for _ in range(nlines)]
        # This statement is very "functional programming"; read it
        # backwards.  List of split lines is "flattened" by chain into
        # iterator of values; this is fed into map to make floats and
        # stored to a list
        return list(map(float, chain(*(line.split() for line in lines))))

    with zopen(bnds_file, "r") as f:
        _ = f.readline()
        if spin_pol:  # Need to read two spin channels
            block_nkpts = int(f.readline().strip()) // 2
            eigenvals = {Spin.up: [], Spin.down: []}
        else:
            block_nkpts = int(f.readline().strip())
            eigenvals = {Spin.up: []}

        while block_nkpts > 0:  # File should be terminated with a 0
            for i in range(block_nkpts):
                kpoint = list(map(float, f.readline().split()))
                kpoints.append(np.array(kpoint) / (alat * _bohr_to_angstrom))

                eigenvals[Spin.up].append(_read_eigenvals(f, eig_lines))

                if spin_pol:
                    spin_down_kpoint = list(map(float, f.readline().split()))
                    if spin_down_kpoint != kpoint:
                        raise AssertionError(
                            "File interpreted as spin-polarised, but this"
                            " kpoint only has one entry: {}".format(kpoint)
                        )
                    eigenvals[Spin.down].append(_read_eigenvals(f, eig_lines))

            block_nkpts = int(f.readline().strip())
            if spin_pol:
                block_nkpts //= 2

    # Transpose matrix to arrange by band and convert to eV from Ry
    eigenvals = {key: np.array(data).T * _ry_to_ev for key, data in eigenvals.items()}
    efermi *= _ry_to_ev

    # Convert labels to Cartesian coordinates
    # [a, b, c] [ax, ay, az] = [a.ax + b.bx + x.cx, a.ay + b.by,...] = [x y z]
    #           |bx, by, bz|
    #           [cx, cy, cz]

    labels = labels or {}  # Initialise a dict if null value
    # (Avoids oddness with mutable option defaults)
    if coords_are_cartesian:
        logging.info(f"Cartesian coordinates, scaling by ALAT = {alat}")
        for label, coords in labels.items():
            labels[label] = np.array(coords) / (alat * _bohr_to_angstrom)
    else:
        for label, coords in labels.items():
            labels[label] = np.dot(
                coords, lattice.reciprocal_lattice_crystallographic.matrix
            )

    # Data in bnds file seems to always be Cartesian
    return BandStructureSymmLine(
        kpoints,
        eigenvals,
        lattice.reciprocal_lattice_crystallographic,
        efermi,
        labels,
        coords_are_cartesian=True,
    )


def dielectric_from_file(filename, out_filename=None):
    """Detect Questaal optics file type and dispatch to appropriate reader

    Args:
        filename (:obj:`str`):
            Path to *opt.ext* or *eps_BSE.out* data file from Questaal. If the
            filename contains "eps_BSE" the *eps_BSE.out* format is assumed,
            otherwise the file is treated as an *opt.ext*.
        out_filename (:obj:`str`, optional):
            Filename for writing out the calculated dielectric data. This is
            only used for *opt.ext* inputs; energy values are converted to eV
            and a Real component is obtained by the Kramers-Kronig relation.
    """

    if "eps_BSE" in filename:
        return dielectric_from_bse(filename)
    else:
        return dielectric_from_opt(filename, out_filename=out_filename)


def dielectric_from_bse(filename):
    """Read a Questaal eps_BSE.out file and return dielectric function

    eps_BSE files only provide a scalar complex number; this is converted to a
    diagonal matrix for compatibility with other routines. Unlike the opt.exe
    however the real part is provided so we do not need to make a
    Kramers-Kronig transformation.

    Args:
        filename (:obj:`float`):
            Path to ``eps_BSE.out`` output of bethesalpeter calculation.

    Returns:
        :obj:`tuple`
            The format imitates the ``dielectric`` attribute of
            :obj:`pymatgen.io.vasp.outputs.Vasprun`: a tuple of the form::

                ([energy1, energy2, ...],
                 [[real_xx_1, real_yy_1, real_zz_1,
                   real_xy_1, real_yz_1, real_xz_1],
                  [real_xx_2, real_yy_2, real_zz_2,
                   real_xy_2, real_yz_2, real_xz_2], ...],
                 [[imag_xx_1, imag_yy_1, imag_zz_1,
                   imag_xy_1, imag_yz_1, imag_xz_1],
                  [imag_xx_2, imag_yy_2, imag_zz_2,
                   imag_xy_2, imag_yz_2, imag_xz_2], ...])

            As only a scalar is given in eps_BSE, the vectors will in fact
            be e.g.::

            [real_1, real_1, real_1, 0, 0, 0]
    """
    data = np.genfromtxt(filename, comments="#")
    if data.shape[1] == 3:
        pass
    else:
        raise ValueError(
            "Not sure how to interpret {}; expected "
            "3 columns. "
            "If this isn't an eps_BSE file, please don't put"
            "eps_BSE in the filename!".format(filename)
        )

    energy = list(data[:, 0].T)
    real = [[r, r, r, 0, 0, 0] for r in data[:, 1]]
    imag = [[i, i, i, 0, 0, 0] for i in data[:, 2]]

    return energy, real, imag


def dielectric_from_opt(filename, cshift=1e-6, out_filename=None):
    """Read a Questaal opt.ext file and return dielectric function

    opt.ext files only provide x, y, z components so the off-diagonal terms are
    set to zero.

    Args:
        filename (:obj:`float`):
            Path to ``opt.ext`` output of LMF optics calculation. If this is
            split into spin channels, the values will be recombined to a single
            (observable) channel.
        cshift (:obj:`float`, optional):
            A small imaginary element used in Kramers-Kronig integration.
        out_filename (:obj:`float`, optional):
            Path to write tabulated dielectric data with columns::

                energy(eV) real_xx real_yy real_zz imag_xx imag_yy imag_zz

            If None, no file is written.

    Returns:
        :obj:`tuple`
            The format imitates the ``dielectric`` attribute of
            :obj:`pymatgen.io.vasp.outputs.Vasprun`: a tuple of the form::

                ([energy1, energy2, ...],
                 [[real_xx_1, real_yy_1, real_zz_1,
                   real_xy_1, real_yz_1, real_xz_1],
                  [real_xx_2, real_yy_2, real_zz_2,
                   real_xy_2, real_yz_2, real_xz_2], ...],
                 [[imag_xx_1, imag_yy_1, imag_zz_1,
                   imag_xy_1, imag_yz_1, imag_xz_1],
                  [imag_xx_2, imag_yy_2, imag_zz_2,
                   imag_xy_2, imag_yz_2, imag_xz_2], ...])

            The off-diagonal (xy, yz, xz) terms are not given in ``opt.ext``
            and are set to zero. Only the imaginary part is provided so the
            real part is constructed by a Kramers-Kronig trasformation.
            Energy units are converted from Ry to eV.
    """

    data = np.genfromtxt(filename, skip_header=1)
    if data.shape[1] == 4:
        pass
    elif data.shape[1] == 7:
        data = np.hstack(data[:, :1], data[1:4] + data[4:7])
    else:
        raise ValueError(
            f"Not sure how to interpret {filename}; expected 4 or 7 columns."
        )

    data[:, 0] *= _ry_to_ev
    de = data[1, 0] - data[0, 0]
    eps_imag = [[[row[1], 0, 0], [0, row[2], 0], [0, 0, row[3]]] for row in data]
    eps_real = kkr(de, eps_imag, cshift=cshift)

    # Re-shape to XX YY ZZ XY YZ XZ format
    eps_imag = np.array(eps_imag).reshape(len(eps_imag), 9)
    eps_imag = eps_imag[:, [0, 4, 8, 1, 5, 2]]
    eps_real = eps_real.reshape(len(eps_real), 9)
    eps_real = eps_real[:, [0, 4, 8, 1, 5, 2]]

    if out_filename is not None:
        with open(out_filename, "wt") as f:
            for e, r, i in zip(data[:, 0].flatten(), eps_real, eps_imag):
                f.write((" ".join(["{:10.8f}"] * 7)).format(e, *r[:3], *i[:3]))
                f.write("\n")

    return data[:, 0].flatten(), eps_real, eps_imag



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
This module provides a class for plotting density of states data.
"""

import itertools

import matplotlib
import matplotlib.pyplot
from matplotlib.ticker import AutoMinorLocator
from pymatgen.electronic_structure.core import Spin

from sumo.electronic_structure.dos import sort_orbitals
from sumo.plotting import (
    colour_cache,
    draw_themed_line,
    pretty_plot,
    pretty_subplot,
    styled_plot,
    sumo_base_style,
    sumo_dos_style,
)

try:
    import configparser
except ImportError:
    import ConfigParser as configparser

empty_space = 1.05


class SDOSPlotter:
    """Class for plotting density of states data.

    This class should be initialised with the total and partial density
    of states. The easiest way to generate the partial density of states is
    using the following method::

        pdos = sumo.electronic_structure.dos.get_pdos()

    Args:
        dos (:obj:`~pymatgen.electronic_structure.dos.Dos`): The total density
            of states.
        pdos (:obj:`dict`, optional): The partial density of states. Formatted
            as a :obj:`dict` of :obj:`dict` mapping the elements and their
            orbitals to :obj:`~pymatgen.electronic_structure.dos.Dos` objects.
            For example::

                {
                    'Bi': {'s': Dos, 'p': Dos ... },
                    'S': {'s': Dos}
                }

            Usually generated using the
            :obj:`sumo.electronic_structure.dos.get_pdos()` function.
    """

    def __init__(self, dos, pdos=None):
        self._dos = dos
        self._pdos = pdos

    def dos_plot_data(
        self,
        yscale=1,
        xmin=-6.0,
        xmax=6.0,
        colours=None,
        plot_total=True,
        legend_cutoff=3,
        subplot=False,
        zero_to_efermi=True,
        zero_energy=None,
        cache=None,
        spin=None,
    ):
        """Get the plotting data.

        Args:
            yscale (:obj:`float`, optional): Scaling factor for the y-axis.
            xmin (:obj:`float`, optional): The minimum energy to mask the
                energy and density of states data (reduces plotting load).
            xmax (:obj:`float`, optional): The maximum energy to mask the
                energy and density of states data (reduces plotting load).
            colours (:obj:`dict`, optional): Use custom colours for specific
                element and orbital combinations. Specified as a :obj:`dict` of
                :obj:`dict` of the colours. For example::

                    {
                        'Sn': {'s': 'r', 'p': 'b'},
                        'O': {'s': '#000000'}
                    }

                The colour can be a hex code, series of rgb value, or any other
                format supported by matplotlib.
            plot_total (:obj:`bool`, optional): Plot the total density of
                states. Defaults to ``True``.
            legend_cutoff (:obj:`float`, optional): The cut-off (in % of the
                maximum density of states within the plotting range) for an
                elemental orbital to be labelled in the legend. This prevents
                the legend from containing labels for orbitals that have very
                little contribution in the plotting range.
            subplot (:obj:`bool`, optional): Plot the density of states for
                each element on separate subplots. Defaults to ``False``.
            zero_to_efermi (:obj:`bool`, optional): Normalise the plot such
                that the Fermi level is set as 0 eV.
            zero_energy (:obj:`float`, optional): If provided, this value is
                used as the 0 eV reference. Otherwise, behaviour depends on
                zero_to_efermi.
            cache (:obj:`dict`, optional): Cache object tracking how colours
                have been assigned to orbitals. The format is the same as the
                "colours" dict. This defaults to the module-level
                sumo.plotting.colour_cache object, but an empty dict can be
                used as a fresh cache. This object will be modified in-place.
            spin (:obj:`Spin`, optional): Check that spin-selection has not
                been called for a closed-shell calculation.

        Returns:
            dict: The plotting data. Formatted with the following keys:

                "energies" (:obj:`numpy.ndarray`)
                    The energies.

                "mask" (:obj:`numpy.ndarray`)
                    A mask used to trim the density of states data and
                    prevent unwanted data being included in the output file.

                "lines" (:obj:`list`)
                    A :obj:`list` of :obj:`dict` containing the density data
                    and some metadata. Each line :obj:`dict` contains the keys:

                        "label" (:obj:`str`)
                            The label for the legend.

                        "dens" (:obj:`numpy.ndarray`)
                            The density of states data.

                        "colour" (:obj:`str`)
                            The colour of the line.

                        "alpha" (:obj:`float`)
                            The alpha value for line fill.

                "ymin" (:obj:`float`)
                    The minimum y-axis limit.

                "ymax" (:obj:`float`)
                    The maximum y-axis limit.
        """
        if cache is None:
            cache = colour_cache

        dos = self._dos
        pdos = self._pdos

        if zero_energy is None:
            energy_shift = -dos.efermi if zero_to_efermi else 0.0
        elif zero_to_efermi:
            raise ValueError(
                "Cannot use zero_energy and zero_to_efermi simultaneously."
            )
        else:
            energy_shift = -zero_energy
        eners = dos.energies + energy_shift

        # mask needed to prevent unwanted data in pdf and for finding y limit
        mask = (eners >= xmin - 0.05) & (eners <= xmax + 0.05)
        plot_data = {"mask": mask, "energies": eners}
        spins = dos.densities.keys()
        if spin is not None and len(spins) == 1:
            raise ValueError(
                "Spin-selection only possible with spin-polarised "
                "calculation results"
            )

        # Visibility cutoff based on scale of total plot even if it is hidden
        try:
            dmax = max(max(d[mask]) for d in dos.densities.values())
        except ValueError:
            dmax = 0

        ymax = dmax if dmax > 0 else 0
        cutoff = (legend_cutoff / 100.0) * (ymax / 1.05)

        if plot_total:
            if "text.color" in matplotlib.rcParams:
                tdos_colour = matplotlib.rcParams["text.color"]
                if tdos_colour is None:
                    tdos_colour = "k"
            else:
                tdos_colour = "k"
            lines = []
            tdos = {
                "label": "Total DOS",
                "dens": dos.densities,
                "colour": tdos_colour,
                "alpha": 0.15,
            }

            # subplot data formatted as a list of lists of dicts, with each
            # list of dicts being plotted on a separate graph, if only one list
            # then solo plot
            lines.append([tdos])

        elif not subplot:
            lines = [[]]  # need a blank list to add lines into
        else:
            lines = []

        for el, el_pdos in pdos.items():
            el_lines = []
            for orb in sort_orbitals(el_pdos):
                dmax = max(max(d[mask]) for d in el_pdos[orb].densities.values())
                ymax = dmax if dmax > ymax else ymax
                label = None if dmax < cutoff else f"{el} ({orb})"
                colour, cache = get_cached_colour(el, orb, colours, cache=cache)
                el_lines.append(
                    {
                        "label": label,
                        "alpha": 0.25,
                        "colour": colour,
                        "dens": el_pdos[orb].densities,
                    }
                )
            if subplot:
                lines.append(el_lines)
            else:
                lines[0].extend(el_lines)

        ymax = ymax * empty_space / yscale
        ymin = 0 if len(spins) == 1 else -ymax
        plot_data.update({"lines": lines, "ymax": ymax, "ymin": ymin})
        return plot_data

    @styled_plot(sumo_base_style, sumo_dos_style)
    def get_plot(
        self,
        subplot=False,
        width=None,
        height=None,
        xmin=-6.0,
        xmax=6.0,
        yscale=1,
        colours=None,
        plot_total=True,
        legend_on=True,
        num_columns=2,
        legend_frame_on=False,
        legend_cutoff=3,
        xlabel="Energy (eV)",
        ylabel="DOS",
        zero_to_efermi=True,
        zero_line=False,
        zero_energy=None,
        dpi=400,
        fonts=None,
        plt=None,
        style=None,
        no_base_style=False,
        spin=None,
    ):
        """Get a :obj:`matplotlib.pyplot` object of the density of states.

        Args:
            subplot (:obj:`bool`, optional): Plot the density of states for
                each element on separate subplots. Defaults to ``False``.
            width (:obj:`float`, optional): The width of the plot.
            height (:obj:`float`, optional): The height of the plot.
            xmin (:obj:`float`, optional): The minimum energy on the x-axis.
            xmax (:obj:`float`, optional): The maximum energy on the x-axis.
            yscale (:obj:`float`, optional): Scaling factor for the y-axis.
            colours (:obj:`dict`, optional): Use custom colours for specific
                element and orbital combinations. Specified as a :obj:`dict` of
                :obj:`dict` of the colours. For example::

                    {
                        'Sn': {'s': 'r', 'p': 'b'},
                        'O': {'s': '#000000'}
                    }

                The colour can be a hex code, series of rgb value, or any other
                format supported by matplotlib.
            plot_total (:obj:`bool`, optional): Plot the total density of
                states. Defaults to ``True``.
            legend_on (:obj:`bool`, optional): Plot the graph legend. Defaults
                to ``True``.
            num_columns (:obj:`int`, optional): The number of columns in the
                legend.
            legend_frame_on (:obj:`bool`, optional): Plot a frame around the
                graph legend. Defaults to ``False``.
            legend_cutoff (:obj:`float`, optional): The cut-off (in % of the
                maximum density of states within the plotting range) for an
                elemental orbital to be labelled in the legend. This prevents
                the legend from containing labels for orbitals that have very
                little contribution in the plotting range.
            xlabel (:obj:`str`, optional): Label/units for x-axis (i.e. energy)
            ylabel (:obj:`str`, optional): Label/units for y-axis (i.e. DOS)
            zero_to_efermi (:obj:`bool`, optional): Normalise the plot such
                that the Fermi level is set as 0 eV.
            zero_line (:obj:`bool`, optional): Draw a line at 0 eV.
            zero_energy (:obj:`float`, optional): If provided, this value is
                used as the 0 eV reference. Otherwise, behaviour depends on
                zero_to_efermi.
            dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
                the image.
            fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
                a single font, specified as a :obj:`str`, or several fonts,
                specified as a :obj:`list` of :obj:`str`.
            plt (:obj:`matplotlib.pyplot`, optional): A
                :obj:`matplotlib.pyplot` object to use for plotting.
            style (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
                style specifications, to be composed on top of Sumo base
                style.
            no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.
            spin (:obj:`Spin`, optional): Plot a spin-polarised density of
                states, "up" or "1" for spin up only, "down" or "-1" for spin
                down only.  Defaults to ``None``.

        Returns:
            :obj:`matplotlib.pyplot`: The density of states plot.
        """
        plot_data = self.dos_plot_data(
            yscale=yscale,
            xmin=xmin,
            xmax=xmax,
            colours=colours,
            plot_total=plot_total,
            legend_cutoff=legend_cutoff,
            subplot=subplot,
            zero_to_efermi=zero_to_efermi,
            zero_energy=zero_energy,
            spin=spin,
        )

        if subplot:
            nplots = len(plot_data["lines"])
            plt = pretty_subplot(
                nplots, 1, width=width, height=height, dpi=dpi, plt=plt
            )
        else:
            plt = pretty_plot(width=width, height=height, dpi=dpi, plt=plt)

        mask = plot_data["mask"]
        energies = plot_data["energies"][mask]
        fig = plt.gcf()
        lines = plot_data["lines"]
        if len(lines[0][0]["dens"]) == 1:
            spins = [Spin.up]
        elif spin is not None:
            spins = [spin]
        else:
            spins = [Spin.up, Spin.down]

        for i, line_set in enumerate(plot_data["lines"]):
            if subplot:
                ax = fig.axes[i]
            else:
                ax = plt.gca()

            for line, spin in itertools.product(line_set, spins):
                if len(spins) == 1:
                    label = line["label"]
                    densities = line["dens"][spin][mask]
                elif spin is Spin.up:
                    label = line["label"]
                    densities = line["dens"][spin][mask]
                elif spin is Spin.down:
                    label = ""
                    densities = -line["dens"][spin][mask]
                ax.fill_between(
                    energies,
                    densities,
                    lw=0,
                    facecolor=line["colour"],
                    alpha=line["alpha"],
                )
                ax.plot(energies, densities, label=label, color=line["colour"])

            if zero_line:
                draw_themed_line(0, ax, orientation="vertical")

            ax.set_ylim(plot_data["ymin"], plot_data["ymax"])
            ax.set_xlim(xmin, xmax)
            if len(spins) == 1:
                ax.set_ylim(0, plot_data["ymax"])
            else:
                ax.set_ylim(plot_data["ymin"], plot_data["ymax"])

            ax.tick_params(axis="y", labelleft=False)
            ax.yaxis.set_minor_locator(AutoMinorLocator(2))
            ax.xaxis.set_minor_locator(AutoMinorLocator(2))

            loc = "upper right" if subplot else "best"
            ncol = 1 if subplot else num_columns
            if legend_on:
                ax.legend(loc=loc, frameon=legend_frame_on, ncol=ncol)

        # no add axis labels and sort out ticks
        if subplot:
            ax.set_xlabel(xlabel)
            fig.subplots_adjust(hspace=0)
            plt.setp([a.get_xticklabels() for a in fig.axes[:-1]], visible=False)
            if "axes.labelcolor" in matplotlib.rcParams:
                ylabelcolor = matplotlib.rcParams["axes.labelcolor"]
            else:
                ylabelcolor = None

            fig.text(
                0.08,
                0.5,
                ylabel,
                ha="left",
                color=ylabelcolor,
                va="center",
                rotation="vertical",
                transform=ax.transAxes,
            )
        else:
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

        return plt


def get_cached_colour(element, orbital, colours=None, cache=None):
    """Get a colour for a particular elemental and orbital combination.

    If the element is not specified in the colours dictionary, the cache is
    checked. If this element-orbital combination has not been chached before,
    a new colour is drawn from the current matplotlib colour cycle and cached.

    The default cache is sumo.plotting.colour_cache. To reset this cache, use
    ``sumo.plotting.colour_cache.clear()``.

    Args:
        element (:obj:`str`): The element.
        orbital (:obj:`str`): The orbital.
        colours (:obj:`dict`, optional): Use custom colours for specific
            element and orbital combinations. Specified as a :obj:`dict` of
            :obj:`dict` of the colours. For example::

                {
                    'Sn': {'s': 'r', 'p': 'b'},
                    'O': {'s': '#000000'}
                }

            The colour can be a hex code, series of rgb value, or any other
            format supported by matplotlib.
        cache (:obj:`dict`, optional): Cache of colour values already
            assigned. The format is the same as the custom colours dict. If
            None, the module-level cache ``sumo.plotting.colour_cache`` is
            used.

    Returns:
        tuple: (colour, cache)
    """

    if cache is None:
        cache = colour_cache

    def _get_colour_with_cache(element, orbital, cache, colour_series):
        """Return cached colour if available, or fetch and cache from cycle"""
        from itertools import chain

        if element in cache and orbital in cache[element]:
            return cache[element][orbital], cache
        else:
            # Iterate through colours to find one which is unused
            for colour in colour_series:
                # Iterate through cache to check if colour already used
                if colour not in chain(
                    *[[col for _, col in orb.items()] for _, orb in cache.items()]
                ):
                    break
            else:
                raise Exception(
                    "Not enough colours available for orbitals! "
                    "Try a different theme."
                )

            if element not in cache:
                cache[element] = {}
            cache[element].update({orbital: colour})
            return colour, cache

    colour_series = matplotlib.rcParams["axes.prop_cycle"].by_key()["color"]

    if isinstance(colours, configparser.ConfigParser):
        try:
            return colours.get(element, orbital), cache
        except (configparser.NoSectionError, configparser.NoOptionError):
            return _get_colour_with_cache(element, orbital, cache, colour_series)

    elif isinstance(colours, dict):
        try:
            return colours[element][orbital], cache
        except KeyError:
            return _get_colour_with_cache(element, orbital, cache, colour_series)

    elif colours is None:
        return _get_colour_with_cache(element, orbital, cache, colour_series)

    else:
        raise TypeError('Argument "colours" should be dict, ' "ConfigParser or None.")



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module providing helper functions for generating k-points along a path.
"""

import logging
import sys


def get_path_data(
    structure,
    mode="bradcrack",
    symprec=0.01,
    spg=None,
    line_density=60,
    cart_coords=False,
    kpt_list=None,
    labels=None,
    phonopy=False,
):
    r"""Get the k-point path, coordinates and symmetry labels for a structure.

    If a manual :obj:`list` of kpoints is supplied using the ``kpt_list``
    variable, the ``mode`` option will be ignored.

    The format of the returned data will be different if phonopy is ``True`` or
    ``False``. This is because phonopy requires the labels and kpoints to be
    provided in a different format than kgen.

    Args:
        structure (:obj:`~pymatgen.core.structure.Structure`): The structure.
        mode (:obj:`str`, optional): Method used for calculating the
            high-symmetry path. The options are:

            bradcrack
                Use the paths from Bradley and Cracknell. See [brad]_.

            pymatgen
                Use the paths from pymatgen. See [curt]_.

            latimer-munro
                Use the paths from Latimer & Munro. See [lm]_.

            seekpath
                Use the paths from SeeK-path. See [seek]_.

        symprec (:obj:`float`, optional): The tolerance for determining the
            crystal symmetry.
        spg (:obj:`~pymatgen.symmetry.groups.SpaceGroup`, optional): Space
            group used to override the symmetry determined by spglib. This is
            not recommended and only provided for testing purposes.
            This option will only take effect when ``mode = 'bradcrack'``.
        line_density (:obj:`int`, optional): Density of k-points along the
            path.
        cart_coords (:obj:`bool`, optional): Whether the k-points are returned
            in cartesian or reciprocal coordinates. Defaults to ``False``
            (fractional coordinates).
        kpt_list (:obj:`list`, optional): List of k-points to use, formatted as
            a list of subpaths, each containing a list of fractional k-points.
            For example::

                [ [[0., 0., 0.], [0., 0., 0.5]],
                  [[0.5, 0., 0.], [0.5, 0.5, 0.]] ]

            Will return points along ``0 0 0 -> 0 0 1/2 | 1/2 0 0
            -> 1/2 1/2 0``
        path_labels (:obj:`list`, optional): The k-point labels. These should
            be provided as a :obj:`list` of :obj:`str` for each subpath of the
            overall path. For example::

                [ ['Gamma', 'Z'], ['X', 'M'] ]

            combined with the above example for ``kpt_list`` would indicate the
            path: Gamma -> Z | X -> M. If no labels are provided, letters from
            A -> Z will be used instead.
        phonopy (:obj:`bool`, optional): Format the k-points and labels for
            use with phonopy. Defaults to ``False``.

    Returns:
        tuple: A tuple of a :obj:`~sumo.symmetry.kpath` object, the k-points
        along the high-symmetry path, and the k-point labels. Returned as
        ``(kpath, kpoints, labels)``.

        The type of ``kpath`` object will depend on the value of ``mode`` and
        whether ``kpt_list`` is set.

        If ``phonopy == False``, then:

            * ``kpoints`` is a :obj:`numpy.ndarray` of the k-point
                coordinates along the high-symmetry path. For example::

                    [[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0], [0.5, 0, 0.25],
                    [0.5, 0, 0.5]]

            * ``labels`` is a :obj:`list` of the high symmetry labels for
                each k-point (will be an empty :obj:`str` if the k-point has
                no label). For example::

                    ['\Gamma', '', 'X', '', 'Y']

        If ``phonopy == True``, then:

            * ``kpoints`` is a :obj:`list` of :obj:`numpy.ndarray`
                containing the k-points for each branch of the band
                structure. This means that the first and last k-points of a
                particular branch may be repeated. For example::

                    [[[0, 0, 0], [0.25, 0, 0], [0.5, 0, 0]],
                    [[0.5, 0, 0], [0.5, 0, 0.25], [0.5, 0, 0.5]]]

            * ``labels`` is a :obj:`list` of the high symmetry labels.
                For example::

                    ['\Gamma', 'X', 'Y']
    """
    from sumo.symmetry import (
        BradCrackKpath,
        CustomKpath,
        LatimerKpath,
        PymatgenKpath,
        SeekpathKpath,
    )

    spg = _get_space_group_object(spg, mode)

    if kpt_list:
        kpath = CustomKpath(structure, kpt_list, labels, symprec=symprec)
    elif mode == "bradcrack":
        kpath = BradCrackKpath(structure, symprec=symprec, spg=spg)
    elif mode == "seekpath":
        kpath = SeekpathKpath(structure, symprec=symprec)
    elif mode == "pymatgen":
        kpath = PymatgenKpath(structure, symprec=symprec)
    elif mode == "latimer-munro":
        kpath = LatimerKpath(structure, symprec=symprec)

    kpoints, labels = kpath.get_kpoints(
        line_density=line_density, phonopy=phonopy, cart_coords=cart_coords
    )
    path_str = kpath.path_string
    kpt_dict = kpath.kpoints

    logging.info("Structure information:")
    logging.info(f"\tSpace group number: {kpath._spg_data['number']}")

    logging.info(f"\tInternational symbol: {kpath.spg_symbol}")
    logging.info(f"\tLattice type: {kpath.lattice_type}")

    logging.info(f"\nk-point path:\n\t{path_str}")
    logging.info("\nk-points:")

    for label, kpoint in iter(kpt_dict.items()):
        coord_str = " ".join([f"{c}" for c in kpoint])
        logging.info(f"\t{label}: {coord_str}")

    return kpath, kpoints, labels


def _get_space_group_object(spg, mode):
    from pymatgen.symmetry.groups import SpaceGroup

    if spg and mode != "bradcrack":
        logging.error(
            "ERROR: Specifying symmetry only supported using "
            "Bradley and Cracknell path."
        )
        sys.exit()
    elif spg:
        try:
            if isinstance(spg, int):
                spg = SpaceGroup.from_int_number(spg)
            else:
                spg = SpaceGroup(spg)
            logging.error(
                "WARNING: Forcing space group not recommended, the "
                "path is likely\nincorrect. Use at your own risk.\n"
            )
        except ValueError:
            logging.error("ERROR: Space group not recognised.")
            sys.exit()
    return spg



import json
import unittest
from os.path import join as path_join

from monty.io import gzip
from monty.json import MontyDecoder
from numpy.testing import assert_array_almost_equal
from pkg_resources import resource_filename
from pymatgen.core.structure import Structure
from pymatgen.electronic_structure.core import Spin

from sumo.io.castep import (
    CastepCell,
    CastepPhonon,
    labels_from_cell,
    read_bands_eigenvalues,
    read_bands_header,
    read_dos,
)

_ry_to_ev = 13.605693009


class CastepCellTestCase(unittest.TestCase):
    def setUp(self):
        self.si_cell = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2.cell")
        )
        self.si_cell_alt = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2-alt.cell")
        )
        self.zns_band_cell = resource_filename(
            __name__, path_join("..", "data", "ZnS", "zns.cell")
        )
        self.zns_singlepoint_cell = resource_filename(
            __name__, path_join("..", "data", "ZnS", "zns-sp.cell")
        )
        si_structure_file = resource_filename(
            __name__, path_join("..", "data", "Si", "Si8.json")
        )
        self.si_structure = Structure.from_file(si_structure_file)

    def test_castep_cell_null_init(self):
        null_cell = CastepCell()
        self.assertEqual(null_cell.blocks, {})
        self.assertEqual(null_cell.tags, {})

        with self.assertRaises(ValueError):
            null_cell.structure

    def test_castep_cell_from_singlepoint_file(self):
        cc = CastepCell.from_file(self.zns_singlepoint_cell)
        self.assertEqual(
            set(cc.blocks.keys()), {"lattice_cart", "positions_abs", "species_pot"}
        )
        self.assertEqual(
            {k: v.value for k, v in cc.tags.items()},
            {
                "fix_all_cell": ["true"],
                "fix_all_ions": ["true"],
                "symmetry_generate": ["true"],
                "kpoint_mp_grid": ["4", "4", "4"],
                "snap_to_symmetry": ["true"],
            },
        )
        self.assertEqual(cc.blocks["species_pot"].values[1], ["S", "NCP"])
        self.assertEqual(cc.blocks["species_pot"].comments, ["", ""])

        structure = cc.structure
        self.assertIsInstance(structure, Structure)
        assert_array_almost_equal(
            structure.lattice.matrix,
            [[0.0, 2.71, 2.71], [2.71, 0.0, 2.71], [2.71, 2.71, 0.0]],
        )

    def test_castep_cell_from_structure(self):
        cell = CastepCell.from_structure(self.si_structure)
        self.assertEqual(
            cell.blocks["lattice_cart"].values,
            [
                ["ang"],
                ["5.43", "0.0", "0.0"],
                ["0.0", "5.43", "0.0"],
                ["0.0", "0.0", "5.43"],
            ],
        )
        self.assertEqual(
            cell.blocks["positions_frac"].values[0], ["Si", "0.0", "0.0", "0.0"]
        )
        self.assertEqual(
            cell.blocks["positions_frac"].values[7], ["Si", "0.75", "0.75", "0.25"]
        )


class CastepDosNiOTestCase(unittest.TestCase):
    def setUp(self):
        nio_files = {
            "bands_file": "NiO.bands",
            "pdos_file": "NiO.pdos_bin",
            "cell_file": "NiO.cell",
        }
        for key, value in nio_files.items():
            nio_files[key] = resource_filename(
                __name__, path_join("..", "data", "NiO", value)
            )
        self.read_dos_kwargs = nio_files

        json_files = {"dos": "dos-pmg.json.gz", "pdos": "pdos-pmg.json.gz"}
        self.ref_data = {}

        for key, value in json_files.items():
            filename = resource_filename(
                __name__, path_join("..", "data", "NiO", value)
            )

            with gzip.open(filename) as f:
                self.ref_data[key] = json.load(f, cls=MontyDecoder)

    def test_tdos_only(self):
        kwargs = self.read_dos_kwargs.copy()
        del kwargs["pdos_file"]
        del kwargs["cell_file"]

        tdos, pdos = read_dos(**kwargs)

        for spin, densities in self.ref_data["dos"].densities.items():
            assert_array_almost_equal(tdos.densities[spin], densities)
        self.assertFalse(pdos)

    def test_missing_cell_file(self):
        """Check error raised if PDOS given without .cell file"""
        kwargs = self.read_dos_kwargs.copy()
        del kwargs["cell_file"]

        with self.assertRaises(OSError):
            tdos, pdos = read_dos(**kwargs)

    def test_pdos(self):
        tdos, pdos = read_dos(**self.read_dos_kwargs)

        for spin, densities in self.ref_data["dos"].densities.items():
            assert_array_almost_equal(tdos.densities[spin], densities)

        for component, orbital_data in self.ref_data["pdos"].items():
            for orbital, dos in orbital_data.items():
                for spin, densities in dos.densities.items():
                    assert_array_almost_equal(
                        pdos[component][orbital].densities[spin], densities
                    )


class CastepBandStructureTestCaseNoSpin(unittest.TestCase):
    def setUp(self):
        self.si_bands = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2.bands")
        )
        self.si_cell = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2.cell")
        )
        self.si_cell_alt = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2-alt.cell")
        )
        self.si_header_ref = resource_filename(
            __name__, path_join("..", "data", "Si", "Si2.bands_header.json")
        )

        self.ref_labels = {
            r"\Gamma": (0.0, 0.0, 0.0),
            "W": (0.5, 0.25, 0.75),
            "L": (0.5, 0.5, 0.5),
            "X": (0.5, 0.0, 0.5),
            "K": (0.375, 0.375, 0.75),
        }

    def test_castep_bands_read_header(self):
        header = read_bands_header(self.si_bands)
        with open(self.si_header_ref) as f:
            ref_header = json.load(f)
        self.assertEqual(header, ref_header)

    def test_castep_bands_read_eigenvalues(self):
        with open(self.si_header_ref) as f:
            ref_header = json.load(f)
        kpoints, weights, eigenvals = read_bands_eigenvalues(self.si_bands, ref_header)

        for i, k in enumerate([0.5, 0.36111111, 0.63888889]):
            self.assertAlmostEqual(kpoints[4][i], k)

        self.assertAlmostEqual(eigenvals[Spin.up][2, 4], 0.09500443 * _ry_to_ev * 2)

        for weight in weights:
            self.assertAlmostEqual(weight, 0.02272727)

    def test_castep_cell_read_labels(self):
        labels = labels_from_cell(self.si_cell)
        self.assertEqual(labels, self.ref_labels)

    def test_castep_cell_read_labels_alt_spelling(self):
        # CASTEP input is case-insensitive anf allows BS_KPOINTS_LIST
        # as well as BS_KPOINT_LIST                            ^
        labels = labels_from_cell(self.si_cell_alt)
        self.assertEqual(labels, self.ref_labels)


class CastepBandStructureTestCaseNickel(unittest.TestCase):
    def setUp(self):
        self.ni_cell = resource_filename(
            __name__, path_join("..", "data", "Ni", "ni-band.cell")
        )

        self.ref_labels = {
            r"\Gamma": (0.0, 0.0, 0.0),
            "L": (0.5, 0.5, 0.5),
            "W": (0.5, 0.25, 0.75),
            "X": (0.5, 0.0, 0.5),
        }

    def test_castep_cell_read_labels_from_list(self):
        # The Si example uses handwritten .cell files in line-mode.
        # This example using a kgen-written .cell file with all k-points listed
        labels = labels_from_cell(self.ni_cell)
        self.assertEqual(labels, self.ref_labels)


class CastepBandStructureTestCaseWithSpin(unittest.TestCase):
    def setUp(self):
        self.fe_bands = resource_filename(
            __name__, path_join("..", "data", "Fe", "Fe.bands")
        )
        self.fe_cell = resource_filename(
            __name__, path_join("..", "data", "Fe", "Fe.cell")
        )
        self.fe_header_ref = resource_filename(
            __name__, path_join("..", "data", "Fe", "Fe.bands_header.json")
        )

    def test_castep_bands_read_header(self):
        header = read_bands_header(self.fe_bands)
        with open(self.fe_header_ref) as f:
            ref_header = json.load(f)
        self.assertEqual(header, ref_header)


# To generate reference file::
#
#   sumo-phonon-bandplot -f zns.phonon --units cm-1 --to-json zns_phonon.json
class CastepPhononTestCaseZincblende(unittest.TestCase):
    def setUp(self):
        self.zns_phonon = resource_filename(
            __name__, path_join("..", "data", "ZnS", "zns.phonon")
        )

        self.zns_cell = resource_filename(
            __name__, path_join("..", "data", "ZnS", "zns.cell")
        )
        self.zns_phonon_ref = resource_filename(
            __name__, path_join("..", "data", "ZnS", "zns_phonon.json")
        )

    def test_castep_phonon_read_bands(self):
        castep_phonon = CastepPhonon.from_file(self.zns_phonon)
        castep_phonon.set_labels_from_file(self.zns_cell)

        bs = castep_phonon.get_band_structure()
        bs_dict = bs.as_dict()

        with open(self.zns_phonon_ref) as f:
            ref_dict = json.load(f)

        for key in bs_dict.keys():
            self.assertIn(key, ref_dict.keys())

        # Comparing this data is quite brittle so compare the arrays when possible
        assert_array_almost_equal(
            bs_dict["lattice_rec"]["matrix"], ref_dict["lattice_rec"]["matrix"]
        )
        assert_array_almost_equal(bs_dict["qpoints"], ref_dict["qpoints"])
        self.assertEqual(bs_dict["labels_dict"], ref_dict["labels_dict"])
        self.assertEqual(bs_dict["structure"]["sites"], ref_dict["structure"]["sites"])
        assert_array_almost_equal(
            bs_dict["structure"]["lattice"]["matrix"],
            ref_dict["structure"]["lattice"]["matrix"],
        )

        assert_array_almost_equal(bs_dict["bands"], ref_dict["bands"])
        assert_array_almost_equal(
            bs_dict["eigendisplacements"]["real"],
            ref_dict["eigendisplacements"]["real"],
        )
        assert_array_almost_equal(
            bs_dict["eigendisplacements"]["imag"],
            ref_dict["eigendisplacements"]["imag"],
        )

        # May be an idea to check has_nac if that logic ever gets sorted out



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
This module provides a class for plotting phonon band structure diagrams.
"""

import itertools
import json
import logging

from matplotlib import rcParams
from matplotlib.cbook import flatten
from matplotlib.ticker import AutoMinorLocator, MaxNLocator
from matplotlib.transforms import blended_transform_factory
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.plotter import PhononBSPlotter

from sumo.plotting import (
    draw_themed_line,
    pretty_plot,
    pretty_subplot,
    styled_plot,
    sumo_base_style,
    sumo_bs_style,
    sumo_phonon_style,
)


class SPhononBSPlotter(PhononBSPlotter):
    """Class for plotting phonon band structures.

    This class is similar to the :obj:`pymatgen.phonon.plotter.PhononBSPlotter`
    class but overrides some methods to generate prettier plots.

    Args:
        bs (:obj:`~pymatgen.phonon.bandstructure.PhononBandStructureSymmLine`):
            The phonon band structure.
    """

    def __init__(self, bs, imag_tol=-5e-2):
        PhononBSPlotter.__init__(self, bs)
        self.imag_tol = imag_tol

    @staticmethod
    def _plot_phonon_dos(dos, ax=None, color=None, dashline=False):
        if ax is None:
            import matplotlib.pyplot as plt

            ax = plt.gca()
        if color is None:
            color = "C0"
        y, x = dos[:, 0], dos[:, 1]
        ax.plot(x, y, "-", color=color)
        ax.fill_betweenx(y, x, 0, color=color, alpha=0.5)
        ax.set_xticks([])
        ax.set_xlim([0, max(x) * 1.1])
        ax.set_xlabel("DOS")

        if dashline:
            draw_themed_line(0, ax)

    @styled_plot(sumo_base_style, sumo_bs_style, sumo_phonon_style)
    def get_plot(
        self,
        units="THz",
        ymin=None,
        ymax=None,
        width=None,
        height=None,
        dpi=None,
        plt=None,
        fonts=None,
        dos=None,
        dos_aspect=3,
        color=None,
        style=None,
        no_base_style=False,
        from_json=None,
        legend=None,
    ):
        """Get a :obj:`matplotlib.pyplot` object of the phonon band structure.

        Args:
            units (:obj:`str`, optional): Units of phonon frequency. Accepted
                (case-insensitive) values are Thz, cm-1, eV, meV.
            ymin (:obj:`float`, optional): The minimum energy on the y-axis.
            ymax (:obj:`float`, optional): The maximum energy on the y-axis.
            width (:obj:`float`, optional): The width of the plot.
            height (:obj:`float`, optional): The height of the plot.
            dpi (:obj:`int`, optional): The dots-per-inch (pixel density) for
                the image.
            fonts (:obj:`list`, optional): Fonts to use in the plot. Can be a
                a single font, specified as a :obj:`str`, or several fonts,
                specified as a :obj:`list` of :obj:`str`.
            plt (:obj:`matplotlib.pyplot`, optional): A
                :obj:`matplotlib.pyplot` object to use for plotting.
            dos (:obj:`np.ndarray`): 2D Numpy array of total DOS data
            dos_aspect (float): Width division for vertical DOS
            color (:obj:`str` or :obj:`tuple`, optional): Line/fill colour in
                any matplotlib-accepted format
            style (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
                style specifications, to be composed on top of Sumo base
                style.
            no_base_style (:obj:`bool`, optional): Prevent use of sumo base
                style. This can make alternative styles behave more
                predictably.
            from_json (:obj:`list` or :obj:`None`, optional): List of paths to
                :obj:`pymatgen.phonon.bandstructure.PhononBandStructureSymmline`
                JSON dump files. These are used to generate additional plots
                displayed under the data attached to this plotter.
                The k-point path should be the same as the main plot; the
                reciprocal lattice is adjusted to fit the scaling of the main
                data input.

        Returns:
            :obj:`matplotlib.pyplot`: The phonon band structure plot.
        """
        if from_json is None:
            from_json = []

        if legend is None:
            legend = [""] * (len(from_json) + 1)
        else:
            if len(legend) == 1 + len(from_json):
                pass
            elif len(legend) == len(from_json):
                legend = [""] + list(legend)
            else:
                raise ValueError("Inappropriate number of legend entries")

        if color is None:
            color = "C0"  # Default to first colour in matplotlib series

        if dos is not None:
            plt = pretty_subplot(
                1,
                2,
                width=width,
                height=height,
                sharex=False,
                sharey=True,
                dpi=dpi,
                plt=plt,
                gridspec_kw={"width_ratios": [dos_aspect, 1], "wspace": 0},
            )
            ax = plt.gcf().axes[0]
        else:
            plt = pretty_plot(width, height, dpi=dpi, plt=plt)
            ax = plt.gca()

        def _plot_lines(data, ax, color=None, alpha=1, zorder=1):
            """Pull data from any PhononBSPlotter and add to axis"""
            dists = data["distances"]
            freqs = data["frequency"]

            # nd is branch index, nb is band index, nk is kpoint index
            for nd, nb in itertools.product(
                range(len(data["distances"])), range(self._nb_bands)
            ):
                f = freqs[nd][nb]

                # plot band data
                ax.plot(dists[nd], f, ls="-", c=color, zorder=zorder)

        data = self.bs_plot_data()
        _plot_lines(data, ax, color=color)

        for i, bs_json in enumerate(from_json):
            with open(bs_json) as f:
                json_data = json.load(f)
                json_data["lattice_rec"] = json.loads(self._bs.lattice_rec.to_json())
                bs = PhononBandStructureSymmLine.from_dict(json_data)

                # bs.lattice_rec = self._bs.lattice_rec
                # raise Exception(bs.qpoints)
            json_plotter = PhononBSPlotter(bs)
            json_data = json_plotter.bs_plot_data()
            if json_plotter._nb_bands != self._nb_bands:
                raise Exception(
                    f"Number of bands in {bs_json} does not match main plot"
                )
            _plot_lines(json_data, ax, color=f"C{i + 1}", zorder=0.5)

        if any(legend):  # Don't show legend if all entries are empty string
            from matplotlib.lines import Line2D

            ax.legend(
                [Line2D([0], [0], color=f"C{i}") for i in range(len(legend))],
                legend,
            )

        self._maketicks(ax, units=units)
        self._makeplot(
            ax,
            plt.gcf(),
            data,
            width=width,
            height=height,
            ymin=ymin,
            ymax=ymax,
            dos=dos,
            color=color,
        )
        plt.tight_layout()
        plt.subplots_adjust(wspace=0)

        return plt

    def _makeplot(
        self,
        ax,
        fig,
        data,
        ymin=None,
        ymax=None,
        height=6,
        width=6,
        dos=None,
        color=None,
    ):
        """Utility method to tidy phonon band structure diagrams."""
        # Define colours
        if color is None:
            color = "C0"  # Default to first colour in matplotlib series

        # set x and y limits
        tymax = ymax if (ymax is not None) else max(flatten(data["frequency"]))
        tymin = ymin if (ymin is not None) else min(flatten(data["frequency"]))
        pad = (tymax - tymin) * 0.05

        if ymin is None:
            ymin = 0 if tymin >= self.imag_tol else tymin - pad
        ymax = ymax if ymax else tymax + pad

        ax.set_ylim(ymin, ymax)
        ax.set_xlim(0, data["distances"][-1][-1])

        if ymin < 0:
            dashline = True
            draw_themed_line(0, ax)

        else:
            dashline = False

        if dos is not None:
            self._plot_phonon_dos(dos, ax=fig.axes[1], color=color, dashline=dashline)
        else:

            # keep correct aspect ratio; match axis to canvas
            x0, x1 = ax.get_xlim()
            y0, y1 = ax.get_ylim()

            if width is None:
                width = rcParams["figure.figsize"][0]
            if height is None:
                height = rcParams["figure.figsize"][1]
            ax.set_aspect((height / width) * ((x1 - x0) / (y1 - y0)))

    def _maketicks(self, ax, units="THz"):
        """Utility method to add tick marks to a band structure."""
        # set y-ticks
        ax.yaxis.set_major_locator(MaxNLocator(6))
        ax.yaxis.set_minor_locator(AutoMinorLocator(2))
        ax.xaxis.set_minor_locator(AutoMinorLocator(2))

        # set x-ticks; only plot the unique tick labels
        ticks = self.get_ticks()
        unique_d = []
        unique_l = []
        if ticks["distance"]:
            temp_ticks = list(zip(ticks["distance"], ticks["label"]))
            unique_d.append(temp_ticks[0][0])
            unique_l.append(temp_ticks[0][1])
            for i in range(1, len(temp_ticks)):
                if unique_l[-1] != temp_ticks[i][1]:
                    unique_d.append(temp_ticks[i][0])
                    unique_l.append(temp_ticks[i][1])

        logging.info("\nLabel positions:")
        for dist, label in list(zip(unique_d, unique_l)):
            logging.info(f"\t{dist:.4f}: {label}")

        ax.set_xticks(unique_d)
        ax.set_xticklabels(unique_l)
        ax.xaxis.grid(True, ls="-")

        trans_xdata_yaxes = blended_transform_factory(ax.transData, ax.transAxes)
        ax.vlines(
            unique_d,
            0,
            1,
            transform=trans_xdata_yaxes,
            colors=rcParams["grid.color"],
            linewidth=rcParams["grid.linewidth"],
        )

        # Use a text hyphen instead of a minus sign because some nice fonts
        # like Whitney don't come with a real minus
        labels = {
            "thz": "THz",
            "cm-1": r"cm$^{\mathrm{-}\mathregular{1}}$",
            "ev": "eV",
            "mev": "meV",
        }
        ax.set_ylabel(f"Frequency ({labels[units.lower()]})")



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Sumo is a set of scripts and an API for dealing with VASP output files.
"""

__version__ = "2.3.4"



import unittest
import warnings
from os.path import join as path_join

import pkg_resources
from pymatgen.core.structure import Structure

from sumo.symmetry.brad_crack_kpath import BradCrackKpath
from sumo.symmetry.seekpath_kpath import SeekpathKpath


class BradCrackKpathTestCase(unittest.TestCase):
    def setUp(self):
        poscar = pkg_resources.resource_filename(
            __name__, path_join("..", "data", "Cs2SnI6", "dos", "vasprun.xml.gz")
        )
        with warnings.catch_warnings():  # Not interested in Pymatgen warnings
            warnings.simplefilter("ignore")
            self.cs_sn_i_structure = Structure.from_file(poscar)

    def test_bravais_assignment(self):
        """Check BradCrack space group logic"""
        self.assertEqual(
            BradCrackKpath._get_bravais_lattice(
                None, "triclinic", None, None, None, None
            ),
            "triclinic",
        )
        self.assertEqual(
            BradCrackKpath._get_bravais_lattice("Fd-3m", "cubic", 5.66, 5.66, 5.66, 0),
            "cubic_f",
        )

    def test_bradcrack_data(self):
        """Check BradCrack data loading"""
        trig_p_c = BradCrackKpath._get_bradcrack_data("trig_p_c")
        self.assertEqual(trig_p_c["path"][0][4], r"\Gamma")
        self.assertEqual(trig_p_c["kpoints"]["K"][1], 0.667)

        mon_p_c = BradCrackKpath._get_bradcrack_data("mon_p_c")
        self.assertEqual(mon_p_c["path"][0][7], "E0")
        self.assertEqual(mon_p_c["kpoints"]["A0"][2], 0.0)

    def test_bradcrack_seekpath_consistent(self):
        """Check BradCrack k-point locations are consistent with Seekpath"""
        kpath_bradcrack = BradCrackKpath(self.cs_sn_i_structure)
        kpath_seekpath = SeekpathKpath(self.cs_sn_i_structure)

        for label, position in kpath_bradcrack.kpoints.items():
            self.assertIn(label, kpath_seekpath.kpoints)
            self.assertEqual(position, kpath_seekpath.kpoints[label])


if __name__ == "__main__":
    unittest.main()



"""
sumo: Heavy weight plotting tools.
"""

from setuptools import find_packages, setup

from sumo import __version__

with open("README.rst") as file:
    long_description = file.read()

setup(
    name="sumo",
    version=__version__,
    description="Heavy weight plotting tools for ab initio solid-state calculations",
    url="https://github.com/smtg-ucl/sumo",
    author="Alex Ganose, Adam J. Jackson",
    author_email="d.scanlon@ucl.ac.uk",
    long_description=long_description,
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords="chemistry pymatgen dft vasp dos band",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        "spglib",
        "numpy",
        "scipy",
        "h5py",
        "pymatgen>=2020.10.20",
        "phonopy>=2.1.3",
        "matplotlib",
        "seekpath",
        "castepxbin<=1.0",
        "colormath",
    ],
    extras_require={
        "docs": ["sphinx", "sphinx-argparse"],
        "tests": ["pytest"],
        "dev": ["pre-commit"],
    },
    package_data={
        "sumo": [
            "symmetry/bradcrack.json",
            "plotting/orbital_colours.conf",
            "plotting/sumo_base.mplstyle",
            "plotting/sumo_bs.mplstyle",
            "plotting/sumo_dos.mplstyle",
            "plotting/sumo_optics.mplstyle",
            "plotting/sumo_phonon.mplstyle",
        ]
    },
    data_files=["examples/orbital_colours.conf", "LICENSE"],
    entry_points={
        "console_scripts": [
            "sumo-bandplot = sumo.cli.bandplot:main",
            "sumo-bandstats = sumo.cli.bandstats:main",
            "sumo-dosplot = sumo.cli.dosplot:main",
            "sumo-kgen = sumo.cli.kgen:main",
            "sumo-phonon-bandplot = sumo.cli.phonon_bandplot:main",
            "sumo-optplot = sumo.cli.optplot:main",
        ]
    },
)



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Module containing helper functions for dealing with
:obj:`~pymatgen.electronic_structure.dos.Dos` and
:obj:`~pymatgen.electronic_structure.dos.CompleteDos` objects.
"""


import logging
import os

import numpy as np
from pymatgen.core.periodic_table import get_el_sp
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.io.vasp.outputs import Vasprun


def load_dos(
    vasprun,
    elements=None,
    lm_orbitals=None,
    atoms=None,
    gaussian=None,
    total_only=False,
    log=False,
    adjust_fermi=True,
    scissor=None,
):
    """Load a vasprun and extract the total and projected density of states.

    Args:
        vasprun (str): Path to a vasprun.xml or vasprun.xml.gz file or
            a :obj:`pymatgen.io.vasp.outputs.Vasprun` object.
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the POSCAR. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.
        gaussian (:obj:`float`, optional): Broaden the density of states using
            convolution with a gaussian function. This parameter controls the
            sigma or standard deviation of the gaussian distribution.
        total_only (:obj:`bool`, optional): Only extract the total density of
            states. Defaults to ``False``.
        log (:obj:`bool`): Print logging messages. Defaults to ``False``.
        adjust_fermi (:obj:`bool`, optional): Shift the Fermi level to sit at
            the valence band maximum (does not affect metals).
        scissor (:obj:`float`, optional): Apply a scissor operator to shift the band gap
            to equal scissor (rigid shift of the densities above the mid gap). Not
            compatible with metallic systems.

    Returns:
        dict: The total and projected density of states. Formatted as a
        :obj:`tuple` of ``(dos, pdos)``, where ``dos`` is a
        :obj:`~pymatgen.electronic_structure.dos.Dos` object containing the
        total density of states and ``pdos`` is a :obj:`dict` of
        :obj:`dict` mapping the elements and their orbitals to
        :obj:`~pymatgen.electronic_structure.dos.Dos` objects. For example::

            {
                'Bi': {'s': Dos, 'p': Dos ... },
                'S': {'s': Dos}
            }
    """
    if isinstance(vasprun, str):
        vr = Vasprun(vasprun)
    else:
        vr = vasprun

    band = vr.get_band_structure()
    dos = vr.complete_dos

    dos, band = _scissor_dos(dos, band, scissor)

    if band.is_metal():
        if log:
            logging.info("System is metallic")
        zero_point = vr.efermi
    else:
        if log:
            logging.info(f"Band gap: {band.get_band_gap()['energy']:.3f}")
            logging.info(f"DOS band gap: {dos.get_gap():.3f}")
        zero_point = band.get_vbm()["energy"]

    if adjust_fermi:
        dos.efermi -= dos.efermi - zero_point

    if gaussian:
        dos.densities = dos.get_smeared_densities(gaussian)
        for site in dos.pdos:
            for orbital in dos.pdos[site]:
                dos.pdos[site][orbital] = dos.get_site_orbital_dos(
                    site, orbital
                ).get_smeared_densities(gaussian)

    if vr.parameters["LSORBIT"]:
        # pymatgen includes the spin down channel for SOC calculations, even
        # though there is no density here. We remove this channel so the
        # plotting is easier later on.
        del dos.densities[Spin.down]
        for site in dos.pdos:
            for orbital in dos.pdos[site]:
                del dos.pdos[site][orbital][Spin.down]

    pdos = {}
    if not total_only:
        pdos = get_pdos(dos, lm_orbitals=lm_orbitals, atoms=atoms, elements=elements)
    return dos, pdos


def get_pdos(dos, lm_orbitals=None, atoms=None, elements=None):
    """Extract the projected density of states from a CompleteDos object.

    Args:
        dos (:obj:`~pymatgen.electronic_structure.dos.CompleteDos`): The
            density of states.
        elements (:obj:`dict`, optional): The elements and orbitals to extract
            from the projected density of states. Should be provided as a
            :obj:`dict` with the keys as the element names and corresponding
            values as a :obj:`tuple` of orbitals. For example, the following
            would extract the Bi s, px, py and d orbitals::

                {'Bi': ('s', 'px', 'py', 'd')}

            If an element is included with an empty :obj:`tuple`, all orbitals
            for that species will be extracted. If ``elements`` is not set or
            set to ``None``, all elements for all species will be extracted.
        lm_orbitals (:obj:`dict`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`dict`, with the elements names as keys and a
            :obj:`tuple` of orbitals as the corresponding values. For example,
            the following would be used to decompose the oxygen p and d
            orbitals::

                {'O': ('p', 'd')}

        atoms (:obj:`dict`, optional): Which atomic sites to use when
            calculating the projected density of states. Should be provided as
            a :obj:`dict`, with the element names as keys and a :obj:`tuple` of
            :obj:`int` specifying the atomic indices as the corresponding
            values. The elemental projected density of states will be summed
            only over the atom indices specified. If an element is included
            with an empty :obj:`tuple`, then all sites for that element will
            be included. The indices are 0 based for each element specified in
            the POSCAR. For example, the following will calculate the density
            of states for the first 4 Sn atoms and all O atoms in the
            structure::

                {'Sn': (1, 2, 3, 4), 'O': (, )}

            If ``atoms`` is not set or set to ``None`` then all atomic sites
            for all elements will be considered.

    Returns:
        dict: The projected density of states. Formatted as a :obj:`dict` of
        :obj:`dict` mapping the elements and their orbitals to
        :obj:`~pymatgen.electronic_structure.dos.Dos` objects. For example::

            {
                'Bi': {'s': Dos, 'p': Dos ... },
                'S': {'s': Dos}
            }
    """
    if not elements:
        symbols = dos.structure.symbol_set
        elements = dict(zip(symbols, [None] * len(symbols)))
    pdos = {}
    for el in elements:
        if atoms and el not in atoms:
            continue

        # select which sites to consider, if no sites were specified then
        # select all. Make a list of the sites of particular elements first
        # due to the dosplot atoms list specification (e.g. starts at 0 for
        # each element
        element_sites = [
            site for site in dos.structure.sites if site.specie == get_el_sp(el)
        ]
        sites = [
            site
            for i, site in enumerate(element_sites)
            if not atoms or (el in atoms and i in atoms[el])
        ]
        lm = lm_orbitals[el] if (lm_orbitals and el in lm_orbitals) else None
        orbitals = elements[el] if elements and el in elements else None

        pdos[el] = get_element_pdos(dos, el, sites, lm, orbitals)
    return pdos


def get_element_pdos(dos, element, sites, lm_orbitals=None, orbitals=None):
    """Get the projected density of states for an element.

    Args:
        dos (:obj:`~pymatgen.electronic_structure.dos.CompleteDos`): The
            density of states.
        element (str): Element symbol. E.g. 'Zn'.
        sites (tuple): The atomic indices over which to sum the density of
            states, as a :obj:`tuple`. Indices are zero based for each
            element. For example, ``(0, 1, 2)`` will sum the density of states
            for the 1st, 2nd and 3rd sites of the element specified.
        lm_orbitals (:obj:`tuple`, optional): The orbitals to decompose into
            their lm contributions (e.g. p -> px, py, pz). Should be provided
            as a :obj:`tuple` of :obj:`str`. For example, ``('p')``, will
            extract the projected density of states for the px, py, and pz
            orbitals. Defaults to ``None``.
        orbitals (:obj:`tuple`, optional): The orbitals to extract from the
            projected density of states. Should be provided as a :obj:`tuple`
            of :obj:`str`. For example, ``('s', 'px', 'dx2')`` will extract the
            s, px, and dx2 orbitals, only. If ``None``, all orbitals will be
            extracted. Defaults to ``None``.

    Returns:
        dict: The projected density of states. Formatted as a :obj:`dict`
        mapping the orbitals to :obj:`~pymatgen.electronic_structure.dos.Dos`
        objects. For example::

            {
                's': Dos,
                'p': Dos
            }
    """
    el_dos = {}
    for site in sites:
        # build a list of which orbitals we are after
        # start with s, p, and d orbitals only
        spd = [
            orb
            for orb in dos.get_element_spd_dos(element).keys()
            if ((orbitals and orb.name in orbitals) or not orbitals)
            and ((lm_orbitals and orb.name not in lm_orbitals) or not lm_orbitals)
        ]

        # now add any lm decomposed orbitals
        lm = [orb for orb in Orbital if lm_orbitals and orb.name[0] in lm_orbitals]

        # extract the data
        for orb in spd:
            pdos = dos.get_site_spd_dos(site)[orb]
            el_dos[orb.name] = el_dos[orb.name] + pdos if orb.name in el_dos else pdos

        for orb in lm:
            pdos = dos.get_site_orbital_dos(site, orb)
            el_dos[orb.name] = el_dos[orb.name] + pdos if orb.name in el_dos else pdos
    return el_dos


def write_files(dos, pdos, prefix=None, directory=None, zero_to_efermi=True):
    """Write the density of states data to disk.

    Args:
        dos (:obj:`~pymatgen.electronic_structure.dos.Dos` or \
             :obj:`~pymatgen.electronic_structure.dos.CompleteDos`): The total
            density of states.
        pdos (dict): The projected density of states. Formatted as a
            :obj:`dict` of :obj:`dict` mapping the elements and their orbitals
            to :obj:`~pymatgen.electronic_structure.dos.Dos` objects. For
            example::

                {
                    'Bi': {'s': Dos, 'p': Dos},
                    'S': {'s': Dos}
                }

        prefix (:obj:`str`, optional): A prefix for file names.
        directory (:obj:`str`, optional): The directory in which to save files.
        zero_to_efermi (:obj:`bool`, optional): Normalise the energy such
             that the Fermi level is set as 0 eV.
    """
    # defining these cryptic lists makes formatting the data much easier later
    if len(dos.densities) == 1:
        sdata = [[Spin.up, 1, ""]]
    else:
        sdata = [[Spin.up, 1, "(up)"], [Spin.down, -1, "(down)"]]

    header = ["energy"]
    eners = dos.energies - dos.efermi if zero_to_efermi else dos.energies
    tdos_data = [eners]
    for spin, sign, label in sdata:
        header.append(f"dos{label}")
        tdos_data.append(dos.densities[spin] * sign)
    tdos_data = np.stack(tdos_data, axis=1)

    filename = f"{prefix}_total_dos.dat" if prefix else "total_dos.dat"
    if directory:
        filename = os.path.join(directory, filename)
    np.savetxt(filename, tdos_data, header=" ".join(header))

    for el, el_pdos in pdos.items():
        header = ["energy"]
        pdos_data = [eners]
        for orb in sort_orbitals(el_pdos):
            for spin, sign, label in sdata:
                header.append(f"{orb}{label}")
                pdos_data.append(el_pdos[orb].densities[spin] * sign)
        pdos_data = np.stack(pdos_data, axis=1)

        if prefix:
            filename = f"{prefix}_{el}_dos.dat"
        else:
            filename = f"{el}_dos.dat"
        if directory:
            filename = os.path.join(directory, filename)
        np.savetxt(filename, pdos_data, header=" ".join(header))


def sort_orbitals(element_pdos):
    """Sort the orbitals of an element's projected density of states.

    Sorts the orbitals based on a standard format. E.g. s < p < d.
    Will also sort lm decomposed orbitals. This is useful for plotting/saving.

    Args:
        element_pdos (dict): An element's pdos. Should be formatted as a
            :obj:`dict` of ``{orbital: dos}``. Where dos is a
            :obj:`~pymatgen.electronic_structure.dos.Dos` object. For example::

                {'s': dos, 'px': dos}

    Returns:
        list: The sorted orbitals.
    """
    sorted_orbitals = [
        "s",
        "p",
        "py",
        "pz",
        "px",
        "d",
        "dxy",
        "dyz",
        "dz2",
        "dxz",
        "dx2",
        "f",
        "f_3",
        "f_2",
        "f_1",
        "f0",
        "f1",
        "f2",
        "f3",
    ]
    unsorted_keys = element_pdos.keys()

    sorted_keys = []
    for key in sorted_orbitals:
        if key in unsorted_keys:
            sorted_keys.append(key)

    return sorted_keys


def _scissor_dos(dos, band, scissor):
    """Apply scissor to DOS by inserting zeros at the midgap point."""
    if scissor and band.is_metal():
        raise ValueError("Cannot apply scissor to metallic DOS.")
    elif scissor:
        mid_gap = (band.get_vbm()["energy"] + band.get_cbm()["energy"]) / 2
        shift = scissor - band.get_band_gap()["energy"]

        # apply scissor to band structure
        for spin, bands in band.bands.items():
            mask = bands > mid_gap
            band.bands[spin][mask] += shift
        band.efermi += shift

        # apply scissor to DOS
        e_step = dos.energies[1] - dos.energies[0]
        n_extra_points = int(np.ceil(shift / e_step))
        dos_fill = np.zeros(n_extra_points)
        energy_fill = np.linspace(
            dos.energies.max(),
            dos.energies.max() + (n_extra_points - 1) * e_step,
            n_extra_points,
        )
        insert_point = np.argmin(np.abs(dos.energies - mid_gap)) + 1

        logging.info(f"Applying scissor to DOS (shift = {shift:.2f} eV)")
        dos.energies = np.concatenate([dos.energies, energy_fill])
        for spin, densities in dos.densities.items():
            dos.densities[spin] = np.insert(densities, insert_point, dos_fill)

        if dos.pdos:
            for site, site_dos in dos.pdos.items():
                for orbital, orbital_dos in site_dos.items():
                    for spin, spin_dos in orbital_dos.items():
                        dos.pdos[site][orbital][spin] = np.insert(
                            spin_dos, insert_point, dos_fill
                        )
    return dos, band



# Copyright (c) Scanlon Materials Theory Group
# Distributed under the terms of the MIT License.

"""
Package containing functions for symmetry and k-point paths.
"""

from sumo.symmetry.kpath import Kpath  # isort:skip
from sumo.symmetry.brad_crack_kpath import BradCrackKpath
from sumo.symmetry.custom_kpath import CustomKpath
from sumo.symmetry.latimer_munro_kpath import LatimerKpath
from sumo.symmetry.pymatgen_kpath import PymatgenKpath
from sumo.symmetry.seekpath_kpath import SeekpathKpath
